# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

import collections
import copy
import datetime

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.tls.algorithm import (
    TlsCipherSuiteExtension,
    TlsECPointFormat,
    TlsGreaseOneByte,
    TlsGreaseTwoByte,
    TlsNamedCurve,
    TlsSignatureAndHashAlgorithm,
)

from cryptoparser.common.exception import InvalidType, NotEnoughData
from cryptoparser.tls.ciphersuite import TlsCipherSuite, SslCipherKind
from cryptoparser.tls.extension import (
    TlsExtensionSupportedVersionsClient,
    TlsExtensionSupportedVersionsServer,
    TlsExtensionUnparsed,
    TlsExtensionEllipticCurves,
    TlsExtensionECPointFormats,
    TlsECPointFormatVector,
    TlsEllipticCurveVector,
    TlsSupportedVersionVector,
)
from cryptoparser.tls.grease import TlsInvalidTypeOneByte, TlsInvalidTypeTwoByte
from cryptoparser.tls.subprotocol import (
    TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM,
    TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM_BYTES,
    SslHandshakeClientHello,
    SslHandshakeServerHello,
    SslMessageType,
    TlsAlertMessage,
    TlsCertificate,
    TlsCertificateStatusType,
    TlsCertificates,
    TlsClientCertificateType,
    TlsCipherSuiteVector,
    TlsCompressionMethod,
    TlsCompressionMethodVector,
    TlsContentType,
    TlsDistinguishedName,
    TlsExtensionType,
    TlsExtensionsClient,
    TlsExtensionsServer,
    TlsHandshakeCertificate,
    TlsHandshakeCertificateRequest,
    TlsHandshakeCertificateStatus,
    TlsHandshakeClientHello,
    TlsHandshakeHelloRandom,
    TlsHandshakeHelloRandomBytes,
    TlsHandshakeHelloRetryRequest,
    TlsHandshakeMessageVariant,
    TlsHandshakeServerHello,
    TlsHandshakeServerHelloDone,
    TlsHandshakeServerKeyExchange,
    TlsHandshakeType,
    TlsSessionIdVector,
    TlsSubprotocolMessageParser,
)

from cryptoparser.tls.record import TlsRecord
from cryptoparser.tls.version import TlsVersion, TlsProtocolVersion

from .classes import TestMessage


class TestSubprotocolParser(unittest.TestCase):
    def test_error(self):
        subprotocol_parser = TlsSubprotocolMessageParser(TlsContentType.HEARTBEAT)
        with self.assertRaises(InvalidValue) as context_manager:
            subprotocol_parser.parse(
                b'\x18' +      # type = heartbeat
                b'\x03\x03' +  # version = TLS 1.2
                b'\x00\x01' +  # length = 1
                b'\x00'
            )
        self.assertEqual(context_manager.exception.value, 0x18)

    def test_registered_parser(self):
        tls_message_dict = collections.OrderedDict([
            ('level', b'\x02'),        # FATAL
            ('description', b'\x28'),  # HANDSHAKE_FAILURE
        ])
        tls_message_bytes = b''.join(tls_message_dict.values())
        tls_parser = TlsSubprotocolMessageParser(TlsContentType.ALERT)
        tls_parser.parse(tls_message_bytes)

        tls_parser.register_subprotocol_parser(TlsContentType.ALERT, TestMessage)
        with self.assertRaises(NotImplementedError):
            tls_parser.parse(tls_message_bytes)

        tls_parser.register_subprotocol_parser(TlsContentType.ALERT, TlsAlertMessage)
        parsed_object, _ = tls_parser.parse(tls_message_bytes)
        self.assertEqual(parsed_object.compose(), tls_message_bytes)


class TestVariantParsable(unittest.TestCase):
    def setUp(self):
        self.server_hello_done_dict = collections.OrderedDict([
            ('handshake_type', b'\x0e'),  # SERVER_HELLO_DONE
            ('length', b'\x00\x00\x00'),  # 0x00
        ])
        self.server_hello_done_bytes = b''.join(self.server_hello_done_dict.values())
        self.server_hello_done = TlsHandshakeServerHelloDone()

    def test_error(self):
        invalid_tls_message_dict = collections.OrderedDict([
            ('content_type', b'\x17'),
            ('data', b'\x00\x00\x00'),
        ])
        invalid_tls_message_bytes = b''.join(invalid_tls_message_dict.values())

        with self.assertRaisesRegex(InvalidValue, 'is not a valid TlsHandshakeMessageVariant'):
            TlsHandshakeMessageVariant.parse_exact_size(invalid_tls_message_bytes)

    def test_compose(self):
        self.assertEqual(TlsHandshakeMessageVariant(self.server_hello_done).compose(), self.server_hello_done_bytes)


class TestTlsCipherSuiteVector(unittest.TestCase):
    def test_parse(self):
        cipher_suites = TlsCipherSuiteVector.parse_exact_size(b'\x00\x02\x00\x00')
        self.assertEqual(cipher_suites, TlsCipherSuiteVector([TlsCipherSuite.TLS_NULL_WITH_NULL_NULL]))

        cipher_suites = TlsCipherSuiteVector.parse_exact_size(b'\x00\x04\x56\x00\x00\xff')
        self.assertEqual(
            cipher_suites,
            TlsCipherSuiteVector([
                TlsInvalidTypeTwoByte(TlsCipherSuiteExtension.FALLBACK_SCSV),
                TlsInvalidTypeTwoByte(TlsCipherSuiteExtension.EMPTY_RENEGOTIATION_INFO_SCSV),
            ])
        )

        cipher_suites = TlsCipherSuiteVector.parse_exact_size(b'\x00\x06\x56\x00\x00\x00\x00\xff')
        self.assertEqual(
            cipher_suites,
            TlsCipherSuiteVector([
                TlsInvalidTypeTwoByte(TlsCipherSuiteExtension.FALLBACK_SCSV),
                TlsCipherSuite.TLS_NULL_WITH_NULL_NULL,
                TlsInvalidTypeTwoByte(TlsCipherSuiteExtension.EMPTY_RENEGOTIATION_INFO_SCSV),
            ])
        )


class TestTlsHandshake(unittest.TestCase):
    def setUp(self):
        self.server_hello_done_dict = collections.OrderedDict([
            ('handshake_type', b'\x0e'),  # SERVER_HELLO_DONE
            ('length', b'\x00\x00\x00'),
        ])
        self.server_hello_done_bytes = b''.join(self.server_hello_done_dict.values())
        self.server_hello_done_record_dict = collections.OrderedDict([
            ('content_type', b'\x16'),          # HANDSHAKE
            ('protocol_version', b'\x03\x01'),  # TLS1
            ('length', b'\x00\x04'),
        ])
        self.server_hello_done_record_bytes = b''.join(self.server_hello_done_record_dict.values())

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            # pylint: disable=expression-not-assigned
            TlsHandshakeClientHello.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, 4)

        with self.assertRaises(InvalidType):
            # pylint: disable=expression-not-assigned
            TlsHandshakeClientHello.parse_exact_size(self.server_hello_done_bytes)

        with self.assertRaises(NotEnoughData) as context_manager:
            TlsHandshakeClientHello.parse_exact_size(
                b'\x01'            # handshake_type: CLIENT_HELLO
                b'\x00\x00\x03' +  # handshake_length = 3
                b'\x03\x03' +      # version = TLS 1.2
                b''
            )
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        with self.assertRaises(InvalidValue) as context_manager:
            TlsHandshakeClientHello.parse_exact_size(
                b'\xff'            # handshake_type: INVALID
                b'\x00\x00\x02' +  # handshake_length = 2
                b'\x03\x03' +      # version = TLS 1.2
                b''
            )
        self.assertEqual(context_manager.exception.value, 0xff)

    def test_parse(self):
        record = TlsRecord.parse_exact_size(
            self.server_hello_done_record_bytes + self.server_hello_done_bytes
        )
        self.assertEqual(record.protocol_version, TlsProtocolVersion(TlsVersion.TLS1))
        record.protocol_version = TlsProtocolVersion(TlsVersion.TLS1_2)
        self.assertEqual(record.protocol_version, TlsProtocolVersion(TlsVersion.TLS1_2))


class TestTlsHandshakeClientHello(unittest.TestCase):
    def setUp(self):
        self.client_hello_minimal_dict = collections.OrderedDict([
            ('handshake_type ', b'\x01'),  # CLIENT_HELLO
            ('length ', b'\x00\x00\x37'),
            ('version ', b'\x03\x03'),
            ('random ',
             b'\x5b\x6c\xd5\x80\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b''),
            ('session_id_length', b'\x00'),
            ('cipher_suite_length', b'\x00\x10'),
            ('cipher_suites',
             b'\x0a\x0a\x00\x01\x00\x02\x00\x03' +
             b'\x00\x04\x00\x05\x56\x00\x00\xff' +
             b''),
            ('compression_method_length', b'\x01'),
            ('compression_methods', b'\x00'),
        ])
        self.client_hello_minimal_bytes = b''.join(self.client_hello_minimal_dict.values())
        self.client_hello_minimal_extensions_dict = collections.OrderedDict([
            ('extensions_length', b'\x00\x0d'),
            ('extension_type', b'\x00\x2b'),  # SUPPORTED_VERSIONS
            ('extension_length', b'\x00\x05'),
            ('supported_version_list_length', b'\x04'),
            ('supported_version_list', b'\x03\x02\x03\x03'),  # TLS1_1, TLS1_2
            ('extension_grease', b'\x0a\x0a'),
            ('extension_grease_length', b'\x00\x00'),
        ])
        self.client_hello_minimal_extensions_bytes = b''.join(self.client_hello_minimal_extensions_dict.values())
        self.client_hello_extension_bytes = bytearray(
            self.client_hello_minimal_bytes +
            self.client_hello_minimal_extensions_bytes +
            b''
        )
        self.client_hello_extension_bytes[3] += (
            len(self.client_hello_extension_bytes) -
            len(self.client_hello_minimal_bytes)
        )

        self.random_time = datetime.datetime(2018, 8, 10, tzinfo=datetime.timezone.utc)
        self.client_hello_minimal = TlsHandshakeClientHello(
            TlsCipherSuiteVector([
                TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A),
                TlsCipherSuite.TLS_RSA_WITH_NULL_MD5,
                TlsCipherSuite.TLS_RSA_WITH_NULL_SHA,
                TlsCipherSuite.TLS_RSA_EXPORT_WITH_RC4_40_MD5,
                TlsCipherSuite.TLS_RSA_WITH_RC4_128_MD5,
                TlsCipherSuite.TLS_RSA_WITH_RC4_128_SHA,
            ]),
            TlsProtocolVersion(TlsVersion.TLS1_2),
            TlsHandshakeHelloRandom(
                self.random_time,
                TlsHandshakeHelloRandomBytes(bytearray(
                    b'\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b''
                ))
            ),
            TlsSessionIdVector(()),
            TlsCompressionMethodVector([TlsCompressionMethod.NULL, ]),
            TlsExtensionsClient(()),
            fallback_scsv=True,
            empty_renegotiation_info_scsv=True,
        )

    def test_parse(self):
        client_hello_minimal = TlsHandshakeClientHello.parse_exact_size(self.client_hello_minimal_bytes)

        self.assertEqual(client_hello_minimal.get_handshake_type(), TlsHandshakeType.CLIENT_HELLO)

        self.assertEqual(
            client_hello_minimal.protocol_version,
            TlsProtocolVersion(TlsVersion.TLS1_2)
        )
        self.assertEqual(
            client_hello_minimal.random,
            self.client_hello_minimal.random
        )
        self.assertEqual(
            client_hello_minimal.random.time,
            self.random_time,
        )
        self.assertEqual(
            client_hello_minimal.random.random,
            self.client_hello_minimal.random.random
        )
        self.assertEqual(
            client_hello_minimal.cipher_suites,
            self.client_hello_minimal.cipher_suites
        )
        self.assertEqual(
            client_hello_minimal.compression_methods,
            self.client_hello_minimal.compression_methods
        )
        self.assertEqual(
            client_hello_minimal.extensions,
            self.client_hello_minimal.extensions
        )
        self.assertTrue(client_hello_minimal.fallback_scsv)
        self.assertTrue(client_hello_minimal.empty_renegotiation_info_scsv)

        client_hello_extension = TlsHandshakeClientHello.parse_exact_size(self.client_hello_extension_bytes)
        self.assertEqual(len(client_hello_extension.extensions), 2)
        self.assertEqual(
            client_hello_extension.extensions.get_item_by_type(TlsExtensionType.SUPPORTED_VERSIONS),
            TlsExtensionSupportedVersionsClient(TlsSupportedVersionVector([
                TlsProtocolVersion(TlsVersion.TLS1_1),
                TlsProtocolVersion(TlsVersion.TLS1_2),
            ]))
        )
        self.assertEqual(
            client_hello_extension.extensions[1],
            TlsExtensionUnparsed(TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A), b'')
        )
        with self.assertRaises(KeyError):
            client_hello_extension.extensions.get_item_by_type(TlsGreaseTwoByte.GREASE_0A0A)

    def test_compose(self):
        self.assertEqual(
            self.client_hello_minimal.compose(),
            self.client_hello_minimal_bytes
        )
        client_hello_extension = TlsHandshakeClientHello.parse_exact_size(self.client_hello_extension_bytes)
        self.assertEqual(
            client_hello_extension.compose(),
            self.client_hello_extension_bytes
        )

    def test_ja3(self):
        client_hello_minimal = copy.copy(self.client_hello_minimal)
        self.assertEqual(client_hello_minimal.ja3(), '771,2570-1-2-3-4-5,,,')

        client_hello_minimal.extensions.append(
            TlsExtensionEllipticCurves(TlsEllipticCurveVector([TlsNamedCurve.SECT163K1]))
        )
        self.assertEqual(client_hello_minimal.ja3(), '771,2570-1-2-3-4-5,10,1,')
        client_hello_minimal.extensions[0].elliptic_curves.append(TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A))
        self.assertEqual(client_hello_minimal.ja3(), '771,2570-1-2-3-4-5,10,1,')

        client_hello_minimal.extensions.append(
            TlsExtensionECPointFormats(TlsECPointFormatVector([TlsECPointFormat.UNCOMPRESSED]))
        )
        self.assertEqual(client_hello_minimal.ja3(), '771,2570-1-2-3-4-5,10-11,1,0')
        client_hello_minimal.extensions[1].point_formats.append(TlsInvalidTypeOneByte(TlsGreaseOneByte.GREASE_0B))
        self.assertEqual(client_hello_minimal.ja3(), '771,2570-1-2-3-4-5,10-11,1,0')


class TestTlsHandshakeServerHello(unittest.TestCase):
    def setUp(self):
        self.server_hello_minimal_dict = collections.OrderedDict([
            ('handshake_type', b'\x02'),              # SERVER_HELLO
            ('length', b'\x00\x00\x26'),
            ('version', b'\x03\x03'),                 # TLS1_2
            ('random',
             b'\x5b\x6c\xd5\x80\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b''),
            ('session_id_length', b'\x00'),
            ('cipher_suite', b'\x00\x01'),
            ('compression_method', b'\x00'),
        ])
        self.server_hello_minimal_bytes = b''.join(self.server_hello_minimal_dict.values())

        self.server_hello_minimal = TlsHandshakeServerHello(
            TlsProtocolVersion(TlsVersion.TLS1_2),
            TlsHandshakeHelloRandom(
                datetime.datetime(2018, 8, 10, tzinfo=datetime.timezone.utc),
                TlsHandshakeHelloRandomBytes(bytearray(
                    b'\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b''
                ))
            ),
            TlsSessionIdVector(()),
            TlsCompressionMethod.NULL,
            TlsCipherSuite.TLS_RSA_WITH_NULL_MD5,
            TlsExtensionsServer(())
        )

        self.server_hello_minimal_bytes = b''.join(self.server_hello_minimal_dict.values())
        self.server_hello_minimal_extensions_dict = collections.OrderedDict([
            ('extensions_length', b'\x00\x0a'),
            ('extension_type', b'\x00\x2b'),  # SUPPORTED_VERSIONS
            ('extension_length', b'\x00\x05'),
            ('selected_version', b'\x03\x03'),  # TLS1_2
            ('extension_grease', b'\x0a\x0a'),
            ('extension_grease_length', b'\x00\x00'),
        ])
        self.server_hello_minimal_extensions_bytes = b''.join(self.server_hello_minimal_extensions_dict.values())
        self.server_hello_extension_bytes = bytearray(
            self.server_hello_minimal_bytes +
            self.server_hello_minimal_extensions_bytes +
            b''
        )
        self.server_hello_extension_bytes[3] += (
            len(self.server_hello_extension_bytes) -
            len(self.server_hello_minimal_bytes)
        )

    def test_parse(self):
        server_hello_minimal = TlsHandshakeServerHello.parse_exact_size(self.server_hello_minimal_bytes)

        self.assertEqual(server_hello_minimal.get_handshake_type(), TlsHandshakeType.SERVER_HELLO)

        self.assertEqual(
            server_hello_minimal.protocol_version,
            TlsProtocolVersion(TlsVersion.TLS1_2)
        )
        self.assertEqual(
            server_hello_minimal.random,
            self.server_hello_minimal.random
        )
        self.assertEqual(
            server_hello_minimal.cipher_suite,
            self.server_hello_minimal.cipher_suite
        )
        self.assertEqual(
            server_hello_minimal.compression_method,
            self.server_hello_minimal.compression_method
        )
        self.assertEqual(
            server_hello_minimal.extensions,
            self.server_hello_minimal.extensions
        )

        server_hello_extension = TlsHandshakeServerHello.parse_exact_size(self.server_hello_extension_bytes)
        self.assertEqual(len(server_hello_extension.extensions), 2)
        self.assertEqual(
            server_hello_extension.extensions.get_item_by_type(TlsExtensionType.SUPPORTED_VERSIONS),
            TlsExtensionSupportedVersionsServer(TlsProtocolVersion(TlsVersion.TLS1_2))
        )
        self.assertEqual(
            server_hello_extension.extensions[1],
            TlsExtensionUnparsed(TlsInvalidTypeTwoByte(TlsGreaseTwoByte.GREASE_0A0A), b'')
        )
        with self.assertRaises(KeyError):
            server_hello_extension.extensions.get_item_by_type(TlsGreaseTwoByte.GREASE_0A0A)

    def test_compose(self):
        self.assertEqual(
            self.server_hello_minimal.compose(),
            self.server_hello_minimal_bytes
        )


class TestTlsHandshakeHelloRetryRequest(unittest.TestCase):
    def setUp(self):
        self.hello_retry_request_minimal_dict = collections.OrderedDict([
            ('handshake_type', b'\x06'),              # HELLO_RETRY_REQUEST
            ('length', b'\x00\x00\x26'),
            ('version', b'\x03\x03'),                 # TLS1_2
            ('random',
             b'\x5b\x6c\xd5\x80\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b''),
            ('session_id_length', b'\x00'),
            ('cipher_suite', b'\x00\x01'),
            ('compression_method', b'\x00'),
        ])
        self.hello_retry_request_minimal_bytes = b''.join(self.hello_retry_request_minimal_dict.values())

        self.hello_retry_request_minimal = TlsHandshakeHelloRetryRequest(
            TlsCipherSuite.TLS_RSA_WITH_NULL_MD5,
            TlsProtocolVersion(TlsVersion.TLS1_2),
            TlsHandshakeHelloRandom(
                datetime.datetime(2018, 8, 10, tzinfo=datetime.timezone.utc),
                TlsHandshakeHelloRandomBytes(bytearray(
                    b'\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                    b''
                ))
            ),
            TlsSessionIdVector(()),
            TlsCompressionMethod.NULL,
            TlsExtensionsClient(())
        )

    def test_parse(self):
        hello_retry_request_minimal = TlsHandshakeHelloRetryRequest.parse_exact_size(
            self.hello_retry_request_minimal_bytes
        )

        self.assertEqual(hello_retry_request_minimal.get_handshake_type(), TlsHandshakeType.HELLO_RETRY_REQUEST)

        self.assertEqual(
            hello_retry_request_minimal.protocol_version,
            TlsProtocolVersion(TlsVersion.TLS1_2)
        )
        self.assertEqual(
            hello_retry_request_minimal.random_bytes,
            self.hello_retry_request_minimal.random_bytes
        )
        self.assertEqual(
            hello_retry_request_minimal.cipher_suite,
            self.hello_retry_request_minimal.cipher_suite
        )
        self.assertEqual(
            hello_retry_request_minimal.compression_method,
            self.hello_retry_request_minimal.compression_method
        )
        self.assertEqual(
            hello_retry_request_minimal.extensions,
            self.hello_retry_request_minimal.extensions
        )

    def test_compose(self):
        self.assertEqual(
            self.hello_retry_request_minimal.compose(),
            self.hello_retry_request_minimal_bytes
        )

    def test_random(self):
        self.assertEqual(
            TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM.compose(),
            TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM_BYTES
        )
        self.assertEqual(
            TlsHandshakeHelloRandom.parse_exact_size(TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM_BYTES),
            TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM
        )


class TestTlsHandshakeCertificate(unittest.TestCase):
    def setUp(self):
        self.certificate_minimal_dict = collections.OrderedDict([
            ('handshake_type', b'\x0b'),                  # CERTIFICATE
            ('length', b'\x00\x00\x31'),
            ('cretificates', b'\x00\x00\x2e'),
            ('peer_cretificate_length', b'\x00\x00\x10'),
            ('peer_certificate_bytes', b'peer certificate'),
            ('intermrdiate_cretificate_length', b'\x00\x00\x18'),
            ('intermrdiate_certificate_bytes', b'intermediate certificate'),
        ])
        self.certificate_minimal_bytes = b''.join(self.certificate_minimal_dict.values())

        self.certificate_minimal = TlsHandshakeCertificate(
            TlsCertificates([
                TlsCertificate(b'peer certificate'),
                TlsCertificate(b'intermediate certificate'),
            ])
        )

    def test_parse(self):
        certificate_minimal = TlsHandshakeCertificate.parse_exact_size(self.certificate_minimal_bytes)

        self.assertEqual(
            certificate_minimal.certificate_chain,
            self.certificate_minimal.certificate_chain
        )

    def test_compose(self):
        self.assertEqual(
            self.certificate_minimal.compose(),
            self.certificate_minimal_bytes
        )


class TestTlsHandshakeCertificateRequestTls10(unittest.TestCase):
    def setUp(self):
        self.certificate_request_dict = collections.OrderedDict([
            ('handshake_type', b'\x0d'),                  # CERTIFICATE_REQUEST
            ('length', b'\x00\x00\x19'),
            ('certificate_types_length', b'\x04'),
            ('certificate_types', b'\x01\02\x03\x04'),
            ('certificate_authorities_length', b'\x00\x12'),
            ('certificate_authority_length', b'\x00\x10'),
            ('certificate_authority', b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
        ])
        self.certificate_request_bytes = b''.join(self.certificate_request_dict.values())

        self.certificate_request = TlsHandshakeCertificateRequest(
            certificate_types=[
                TlsClientCertificateType.RSA_SIGN,
                TlsClientCertificateType.DSS_SIGN,
                TlsClientCertificateType.RSA_FIXED_DH,
                TlsClientCertificateType.DSS_FIXED_DH,
            ],
            certificate_authorities=[
                TlsDistinguishedName(b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
            ]
        )

    def test_parse(self):
        certificate_request = TlsHandshakeCertificateRequest.parse_exact_size(self.certificate_request_bytes)

        self.assertEqual(
            list(certificate_request.certificate_types),
            [
                TlsClientCertificateType.RSA_SIGN,
                TlsClientCertificateType.DSS_SIGN,
                TlsClientCertificateType.RSA_FIXED_DH,
                TlsClientCertificateType.DSS_FIXED_DH,
            ]
        )
        self.assertEqual(
            list(certificate_request.certificate_authorities),
            [TlsDistinguishedName(b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'), ]
        )
        self.assertEqual(certificate_request.supported_signature_algorithms, None)

    def test_compose(self):
        self.assertEqual(
            self.certificate_request.compose(),
            self.certificate_request_bytes
        )


class TestTlsHandshakeCertificateRequestTls12(unittest.TestCase):
    def setUp(self):
        self.certificate_request_dict = collections.OrderedDict([
            ('handshake_type', b'\x0d'),                  # CERTIFICATE_REQUEST
            ('length', b'\x00\x00\x23'),
            ('certificate_types_length', b'\x04'),
            ('certificate_types', b'\x01\02\x03\x04'),
            ('signature_algorithm_list_length', b'\x00\x08'),
            ('signature_algorithm_list', b'\x01\x00\x02\x01\x03\x02\x04\x03'),
            ('certificate_authorities_length', b'\x00\x12'),
            ('certificate_authority_length', b'\x00\x10'),
            ('certificate_authority', b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
        ])
        self.certificate_request_bytes = b''.join(self.certificate_request_dict.values())

        self.certificate_request = TlsHandshakeCertificateRequest(
            certificate_types=[
                TlsClientCertificateType.RSA_SIGN,
                TlsClientCertificateType.DSS_SIGN,
                TlsClientCertificateType.RSA_FIXED_DH,
                TlsClientCertificateType.DSS_FIXED_DH,
            ],
            supported_signature_algorithms=[
                TlsSignatureAndHashAlgorithm.ANONYMOUS_MD5,
                TlsSignatureAndHashAlgorithm.RSA_SHA1,
                TlsSignatureAndHashAlgorithm.DSA_SHA224,
                TlsSignatureAndHashAlgorithm.ECDSA_SHA256,
            ],
            certificate_authorities=[
                TlsDistinguishedName(b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
            ]
        )

    def test_parse(self):
        certificate_request = TlsHandshakeCertificateRequest.parse_exact_size(self.certificate_request_bytes)

        self.assertEqual(
            list(certificate_request.certificate_types),
            [
                TlsClientCertificateType.RSA_SIGN,
                TlsClientCertificateType.DSS_SIGN,
                TlsClientCertificateType.RSA_FIXED_DH,
                TlsClientCertificateType.DSS_FIXED_DH,
            ]
        )
        self.assertEqual(
            list(certificate_request.supported_signature_algorithms),
            [
                TlsSignatureAndHashAlgorithm.ANONYMOUS_MD5,
                TlsSignatureAndHashAlgorithm.RSA_SHA1,
                TlsSignatureAndHashAlgorithm.DSA_SHA224,
                TlsSignatureAndHashAlgorithm.ECDSA_SHA256,
            ]
        )
        self.assertEqual(
            list(certificate_request.certificate_authorities),
            [TlsDistinguishedName(b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'), ]
        )

    def test_compose(self):
        self.assertEqual(
            self.certificate_request.compose(),
            self.certificate_request_bytes
        )


class TestTlsHandshakeServerHelloDone(unittest.TestCase):
    def setUp(self):
        self.server_hello_done_dict = collections.OrderedDict([
            ('handshake_type', b'\x0e'),  # SERVER_HELLO_DONE
            ('length', b'\x00\x00\x00'),  # 0x00
        ])
        self.server_hello_done_bytes = b''.join(self.server_hello_done_dict.values())

        self.server_hello_done = TlsHandshakeServerHelloDone()

    def test_error(self):
        error_regex = 'b\'\\\\x00\' is not a valid TlsHandshakeServerHelloDone payload value'
        with self.assertRaisesRegex(InvalidValue, error_regex):
            # pylint: disable=expression-not-assigned
            TlsHandshakeServerHelloDone.parse_exact_size(b'\x0e\x00\x00\x01\x00')

    def test_parse(self):
        server_hello_done = TlsHandshakeServerHelloDone.parse_exact_size(self.server_hello_done_bytes)

        self.assertEqual(server_hello_done.get_handshake_type(), TlsHandshakeType.SERVER_HELLO_DONE)

    def test_compose(self):
        self.assertEqual(self.server_hello_done.compose(), self.server_hello_done_bytes)


class TestTlsHandshakeServerKeyExcahnge(unittest.TestCase):
    def setUp(self):
        self.param_bytes = b'\x00\x01\x02\x03\x04\x05\x06\x07'
        self.server_key_exchange_dict = collections.OrderedDict([
            ('handshake_type', b'\x0c'),       # SERVER_KEY_EXCHANGE
            ('length', b'\x00\x00\x08'),
            ('param_bytes', self.param_bytes),
        ])
        self.server_key_exchange_bytes = b''.join(self.server_key_exchange_dict.values())

        self.server_key_exchange = TlsHandshakeServerKeyExchange(self.param_bytes)

    def test_parse(self):
        server_key_exchange = TlsHandshakeServerKeyExchange.parse_exact_size(self.server_key_exchange_bytes)

        self.assertEqual(server_key_exchange.get_handshake_type(), TlsHandshakeType.SERVER_KEY_EXCHANGE)

        self.assertEqual(server_key_exchange.param_bytes, self.param_bytes)

    def test_compose(self):
        self.assertEqual(self.server_key_exchange.compose(), self.server_key_exchange_bytes)


class TestTlsHandshakeCertificateStatus(unittest.TestCase):
    def setUp(self):
        self.status_bytes = b'\x00\x01\x02\x03\x04\x05\x06\x07'
        self.certificate_status_bytes = bytes(
            b'\x16' +                              # handshake_type = CERTIFICATE_STATUS
            b'\x00\x00\x0c' +                      # length = 0x0c
            b'\x01' +                              # status_type = OCSP
            b'\x00\x00\x08' +                      # length = 0x08
            self.status_bytes +                    # status_bytes
            b''
        )

        self.certificate_status = TlsHandshakeCertificateStatus(TlsCertificateStatusType.OCSP, self.status_bytes)

    def test_parse(self):
        certificate_status = TlsHandshakeCertificateStatus.parse_exact_size(self.certificate_status_bytes)

        self.assertEqual(certificate_status.get_handshake_type(), TlsHandshakeType.CERTIFICATE_STATUS)

        self.assertEqual(certificate_status.status, self.status_bytes)

    def test_compose(self):
        self.assertEqual(self.certificate_status.compose(), self.certificate_status_bytes)


class TestSslHandshakeClientHello(unittest.TestCase):
    def setUp(self):
        self.client_hello_dict = collections.OrderedDict([
            ('version', b'\x00\x02'),                              # SSL2
            ('cipher_kinds_length', b'\x00\x06'),
            ('session_id_length', b'\x00\x08'),
            ('challenge_length', b'\x00\x10'),
            ('cipher_kinds', b'\x01\x00\x80\x07\x00\xc0'),
            ('session_id', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('challenge',
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
             b'')
        ])
        self.client_hello_bytes = b''.join(self.client_hello_dict.values())

        self.client_hello = SslHandshakeClientHello(
            cipher_kinds=[
                SslCipherKind.SSL_CK_RC4_128_WITH_MD5,
                SslCipherKind.SSL_CK_DES_192_EDE3_CBC_WITH_MD5
            ],
            session_id=b'\x00\x01\x02\x03\x04\x05\x06\x07',
            challenge=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        )

    def test_default(self):
        self.assertEqual(SslHandshakeClientHello([]).session_id, b'')
        self.assertNotEqual(SslHandshakeClientHello([]).challenge, SslHandshakeClientHello([]).challenge)

        self.assertEqual(SslHandshakeServerHello(b'', []).connection_id, b'')

    def test_parse(self):
        client_hello_minimal = SslHandshakeClientHello.parse_exact_size(self.client_hello_bytes)

        self.assertEqual(client_hello_minimal.get_message_type(), SslMessageType.CLIENT_HELLO)

    def test_compose(self):
        self.assertEqual(self.client_hello.compose(), self.client_hello_bytes)


class TestSslHandshakeServerHello(unittest.TestCase):
    def setUp(self):
        self.server_hello_done_dict = collections.OrderedDict([
            ('session_id_hit', b'\x00'),                    # False
            ('certificate_type', b'\x01'),                  # X509_CERTIFICATE
            ('version', b'\x00\x02'),                       # SSL2
            ('certificate_length', b'\x00\x0b'),
            ('cipher_kinds_length', b'\x00\x06'),
            ('connection_id_length', b'\x00\x10'),
            ('certificate', b'certificate'),
            ('cipher_kinds', b'\x01\x00\x80\x07\x00\xc0'),
            ('connection_id',
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
             b''),
        ])
        self.server_hello_bytes = b''.join(self.server_hello_done_dict.values())

        self.server_hello = SslHandshakeServerHello(
            certificate=b'certificate',
            cipher_kinds=[
                SslCipherKind.SSL_CK_RC4_128_WITH_MD5,
                SslCipherKind.SSL_CK_DES_192_EDE3_CBC_WITH_MD5
            ],
            connection_id=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
            session_id_hit=False
        )

    def test_parse(self):
        server_hello_minimal = SslHandshakeServerHello.parse_exact_size(self.server_hello_bytes)

        self.assertEqual(server_hello_minimal.get_message_type(), SslMessageType.SERVER_HELLO)

    def test_compose(self):
        self.assertEqual(self.server_hello.compose(), self.server_hello_bytes)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-


from cryptodatahub.common.exception import InvalidValue
from cryptoparser.common.parse import ParsableBase, ParserText, ComposerText


class LanguageTag(ParsableBase):
    def __init__(self, primary_subtag, subsequent_subtags=()):
        self._primary_subtag = None
        self._subsequent_subtags = None

        self.primary_subtag = primary_subtag
        self.subsequent_subtags = subsequent_subtags

    @property
    def primary_subtag(self):
        return self._primary_subtag

    @primary_subtag.setter
    def primary_subtag(self, value):
        if not value or len(value) > 8 or not value.isalpha():
            raise InvalidValue(value, LanguageTag, 'primary_subtag')

        self._primary_subtag = value

    @property
    def subsequent_subtags(self):
        return self._subsequent_subtags

    @subsequent_subtags.setter
    def subsequent_subtags(self, value):
        for subsequent_subtag in value:
            if not subsequent_subtag or len(subsequent_subtag) > 8 or not subsequent_subtag.isalnum():
                raise InvalidValue(value, LanguageTag, 'subsequent_subtag')

        self._subsequent_subtags = value

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)
        parser.parse_string_array('tags', '-')

        return LanguageTag(parser['tags'][0], parser['tags'][1:]), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.primary_subtag)
        if self.subsequent_subtags:
            composer.compose_separator('-')
            composer.compose_string_array(self.subsequent_subtags, '-')

        return composer.composed



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import datetime
import unittest
from unittest import mock

import asn1crypto.x509

from cryptodatahub.common.exception import InvalidValue
from cryptoparser.common.exception import NotEnoughData, TooMuchData, InvalidType
from cryptoparser.common.parse import ParserBinary, ParserText, ParsableBase, ComposerBinary, ComposerText, ByteOrder
from cryptoparser.tls.ciphersuite import TlsCipherSuiteFactory

from .classes import (
    AlwaysInvalidTypeVariantParsable,
    AlwaysTestStringComposer,
    ConditionalParsable,
    FlagEnum,
    OneByteOddParsable,
    OneByteParsable,
    SerializableEnum,
    SerializableEnumFactory,
    SerializableEnumVariantParsable,
    StringEnum,
    TwoByteParsable,
)


class TestParsableBase(unittest.TestCase):
    _ALPHA_BETA_GAMMA = 'αβγ'
    _ALPHA_BETA_GAMMA_BYTES = 'αβγ'.encode('utf-8')
    _ALPHA_BETA_GAMMA_LEN_BYTES = bytes((len(_ALPHA_BETA_GAMMA_BYTES),))

    _ALPHA_BETA_GAMMA_HASHMARK_BYTES = 'αβγ#'.encode('utf-8')


class TestParsable(TestParsableBase):
    def test_error(self):
        with self.assertRaises(TooMuchData) as context_manager:
            OneByteParsable.parse_exact_size(b'\x01\x02')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        with self.assertRaises(NotEnoughData) as context_manager:
            OneByteParsable.parse_immutable(b'')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        with self.assertRaises(TypeError):
            # pylint: disable=protected-access,abstract-class-instantiated
            ParsableBase()._parse(b'')

        with self.assertRaises(TypeError):
            # pylint: disable=abstract-class-instantiated
            ParsableBase().compose()

        parser = ParserBinary(b'\xff\xff')
        with self.assertRaises(InvalidValue):
            parser.parse_parsable('cipher_suite', TlsCipherSuiteFactory)

        with self.assertRaises(InvalidValue) as context_manager:
            AlwaysInvalidTypeVariantParsable.parse_exact_size(b'\x01\x02\x03\x04')
        self.assertEqual(context_manager.exception.value, b'\x01\x02\x03\x04')

        with self.assertRaises(InvalidValue) as context_manager:
            AlwaysInvalidTypeVariantParsable(0)
        self.assertEqual(context_manager.exception.value, 0)

    def test_parse(self):
        _, parsed_length = OneByteParsable.parse_immutable(b'\x01\x02')
        self.assertEqual(parsed_length, 1)

        parsable = bytearray([0x01, 0x02])
        OneByteParsable.parse_mutable(parsable)
        self.assertEqual(parsable, b'\x02')

        parsed_value, parsed_length = SerializableEnumFactory.parse_immutable(b'\x00\x01')
        self.assertEqual(parsed_value, SerializableEnum.FIRST)
        self.assertEqual(parsed_length, 2)

    def test_repr(self):
        self.assertEqual(repr(SerializableEnum.FIRST), 'SerializableEnum.FIRST')
        AlwaysInvalidTypeVariantParsable.register_variant_parser(SerializableEnumFactory, SerializableEnumFactory)


class TestParserBase(TestParsableBase):
    def test_mapping(self):
        parser = ParserBinary(b'\x01\x02')
        self.assertEqual(len(parser), 0)
        self.assertEqual(parser.parsed_length, 0)
        self.assertEqual(parser.unparsed_length, 2)
        self.assertEqual(dict(parser), {})

        parser.parse_numeric('first_byte', 1)
        self.assertEqual(len(parser), 1)
        self.assertEqual(parser.parsed_length, 1)
        self.assertEqual(parser.unparsed_length, 1)
        self.assertEqual(dict(parser), {'first_byte': 1})

        parser.parse_numeric('second_byte', 1)
        self.assertEqual(len(parser), 2)
        self.assertEqual(parser.parsed_length, 2)
        self.assertEqual(parser.unparsed_length, 0)
        self.assertEqual(dict(parser), {'first_byte': 1, 'second_byte': 2})


class TestParserBinary(TestParsableBase):
    def test_error(self):
        parser = ParserBinary(b'\x00')
        parser.parse_numeric('one_byte', 1)
        with self.assertRaises(NotEnoughData) as context_manager:
            parser.parse_numeric('one_byte', 1)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        parser = ParserBinary(b'\x00\x00\x00\x00')
        parser.parse_numeric('one_byte', 1)
        with self.assertRaises(NotEnoughData) as context_manager:
            parser.parse_numeric_array('four_byte_array', item_num=2, item_size=3)
        self.assertEqual(context_manager.exception.bytes_needed, 3)

        parser = ParserBinary(b'\x00\x00\x00\x00')
        parser.parse_numeric('one_byte', 1)
        with self.assertRaises(NotEnoughData) as context_manager:
            parser.parse_bytes('four_byte_array', 4)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        parser = ParserBinary(b'\x01\xff')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_bytes('one_byte_array', 1, converter=mock.Mock(name='mock', side_effect=ValueError))
        self.assertEqual(context_manager.exception.value, 1)

        parser = ParserBinary(b'\x00\x00\x00\x00\x00')
        with self.assertRaises(NotImplementedError):
            parser.parse_numeric('five_byte_numeric', 5)

        parser = ParserBinary(b'\xff\xff')
        with self.assertRaises(InvalidValue):
            parser.parse_numeric('two_byte_numeric', 2, OneByteParsable)

        parser = ParserBinary(b'\x10')
        with self.assertRaises(InvalidValue):
            parser.parse_numeric('flags', 1, FlagEnum)

        with self.assertRaises(InvalidValue) as context_manager:
            AlwaysInvalidTypeVariantParsable.parse_immutable(b'\x00\x00')

        AlwaysInvalidTypeVariantParsable.register_variant_parser(SerializableEnumFactory, SerializableEnumFactory)
        with self.assertRaises(InvalidValue) as context_manager:
            AlwaysInvalidTypeVariantParsable.parse_exact_size(b'\x01\x02')
        self.assertEqual(context_manager.exception.value, 0x0102)

    def test_parse_numeric(self):
        parser = ParserBinary(b'\x01\x02')
        parser.parse_numeric('first_byte', 1)
        parser.parse_numeric('second_byte', 1)
        self.assertEqual(parser['first_byte'], 0x01)
        self.assertEqual(parser['second_byte'], 0x02)

        parser = ParserBinary(b'\x01\x02')
        parser.parse_numeric('first_two_bytes', 2)
        self.assertEqual(parser['first_two_bytes'], 0x0102)

        parser = ParserBinary(b'\x01\x02\x03')
        parser.parse_numeric('first_two_bytes', 3)
        self.assertEqual(parser['first_two_bytes'], 0x010203)

        parser = ParserBinary(b'\x01\x02\x03\x04')
        parser.parse_numeric('first_four_bytes', 4)
        self.assertEqual(parser['first_four_bytes'], 0x01020304)

    def test_parse_byte_order(self):
        parser = ParserBinary(b'\x01\x02\x03\x04', byte_order=ByteOrder.BIG_ENDIAN)
        parser.parse_numeric('number', 4)
        self.assertEqual(parser['number'], 0x01020304)

        parser = ParserBinary(b'\x01\x02\x03\x04', byte_order=ByteOrder.LITTLE_ENDIAN)
        parser.parse_numeric('number', 4)
        self.assertEqual(parser['number'], 0x04030201)

        parser = ParserBinary(b'\x01\x02\x03\x04', byte_order=ByteOrder.NETWORK)
        parser.parse_numeric('number', 4)
        self.assertEqual(parser['number'], 0x01020304)

        parser = ParserBinary(b'\x01\x02\x03', byte_order=ByteOrder.BIG_ENDIAN)
        parser.parse_numeric('number', 3)
        self.assertEqual(parser['number'], 0x010203)

        parser = ParserBinary(b'\x01\x02\x03', byte_order=ByteOrder.LITTLE_ENDIAN)
        parser.parse_numeric('number', 3)
        self.assertEqual(parser['number'], 0x030201)

        parser = ParserBinary(b'\x01\x02\x03', byte_order=ByteOrder.NETWORK)
        parser.parse_numeric('number', 3)
        self.assertEqual(parser['number'], 0x010203)

    def test_parse_numeric_flags(self):
        parser = ParserBinary(b'\x01')
        parser.parse_numeric_flags('flags', 1, FlagEnum)
        self.assertEqual(parser['flags'], {FlagEnum.ONE, })

        parser = ParserBinary(b'\x03')
        parser.parse_numeric_flags('flags', 1, FlagEnum)
        self.assertEqual(parser['flags'], {FlagEnum.ONE, FlagEnum.TWO})

    def test_parse_mpint(self):
        parser = ParserBinary(b'\x00\x00\x00\x00')
        parser.parse_mpint('mpint', 4)
        self.assertEqual(parser['mpint'], 0)

        parser = ParserBinary(b'\x09\xa3\x78\xf9\xb2\xe3\x32\xa7')
        parser.parse_mpint('mpint', 8)
        self.assertEqual(parser['mpint'], 0x9a378f9b2e332a7)

        parser = ParserBinary(b'\x00\x80')
        parser.parse_mpint('mpint', 2)
        self.assertEqual(parser['mpint'], 0x80)

        parser = ParserBinary(b'\xed\xcc')
        parser.parse_mpint('mpint', 2)
        self.assertEqual(parser['mpint'], 0xedcc)

        parser = ParserBinary(b'\xff\x21\x52\x41\x11')
        parser.parse_mpint('mpint', 5)
        self.assertEqual(parser['mpint'], 0xff21524111)

    def test_parse_ssh_mpint(self):
        parser = ParserBinary(b'\x00')
        with self.assertRaises(NotEnoughData) as context_manager:
            parser.parse_ssh_mpint('mpint')
        self.assertEqual(context_manager.exception.bytes_needed, 3)

        parser = ParserBinary(b'\x00\x00\x00\x00')
        parser.parse_ssh_mpint('mpint')
        self.assertEqual(parser['mpint'], 0)

        parser = ParserBinary(b'\x00\x00\x00\x08\x09\xa3\x78\xf9\xb2\xe3\x32\xa7')
        parser.parse_ssh_mpint('mpint')
        self.assertEqual(parser['mpint'], 0x9a378f9b2e332a7)

        parser = ParserBinary(b'\x00\x00\x00\x02\x00\x80')
        parser.parse_ssh_mpint('mpint')
        self.assertEqual(parser['mpint'], 0x80)

        parser = ParserBinary(b'\x00\x00\x00\x02\xed\xcc')
        parser.parse_ssh_mpint('mpint')
        self.assertEqual(parser['mpint'], -0x1234)

        parser = ParserBinary(b'\x00\x00\x00\x05\xff\x21\x52\x41\x11')
        parser.parse_ssh_mpint('mpint')
        self.assertEqual(parser['mpint'], -0xdeadbeef)

    def test_parse_numeric_array(self):
        parser = ParserBinary(b'\x01\x02')
        parser.parse_numeric_array('one_byte_array', item_num=2, item_size=1)
        self.assertEqual(parser['one_byte_array'], [1, 2])

        parser = ParserBinary(b'\x00\x01\x00\x02')
        parser.parse_numeric_array('two_byte_array', item_num=2, item_size=2)
        self.assertEqual(parser['two_byte_array'], [1, 2])

        parser = ParserBinary(b'\x00\x00\x01\x00\x00\x02')
        parser.parse_numeric_array('three_byte_array', item_num=2, item_size=3)
        self.assertEqual(parser['three_byte_array'], [1, 2])

        parser = ParserBinary(b'\x00\x00\x00\x01\x00\x00\x00\x02')
        parser.parse_numeric_array('four_byte_array', item_num=2, item_size=4)
        self.assertEqual(parser['four_byte_array'], [1, 2])

    def test_parse_byte_array(self):
        parser = ParserBinary(b'\x01\x02')
        parser.parse_raw('two_byte_array', size=2)
        self.assertEqual(parser['two_byte_array'], b'\x01\x02')

        parser = ParserBinary(b'not X.509 certificate')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_raw('name', 21, asn1crypto.x509.Certificate.load)
        self.assertEqual(context_manager.exception.value, b'not X.509 certificate')

    def test_parse_string(self):
        parser = ParserBinary(b'\x02\xff\xff')
        with self.assertRaises(InvalidValue):
            parser.parse_string('non-utf-8-string', 1, 'utf-8')

        parser = ParserBinary(self._ALPHA_BETA_GAMMA_LEN_BYTES + self._ALPHA_BETA_GAMMA_BYTES)
        parser.parse_string('utf-8-string', 1, 'utf-8')
        self.assertEqual(parser['utf-8-string'], self._ALPHA_BETA_GAMMA)

    def test_parse_string_null_terminated(self):
        parser = ParserBinary(b'non-null-terminated-string')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_null_terminated('non-utf-8-string', 1, 'utf-8')
        self.assertEqual(context_manager.exception.value, b'non-null-terminated-string')

        parser = ParserBinary(b'1\x00')
        parser.parse_string_null_terminated('one-byte-string', 'utf-8', int)
        self.assertEqual(parser['one-byte-string'], 1)

        parser = ParserBinary(self._ALPHA_BETA_GAMMA_BYTES + b'\x00remaining-data')
        parser.parse_string_null_terminated('utf-8-string', 'utf-8')
        self.assertEqual(parser['utf-8-string'], self._ALPHA_BETA_GAMMA)
        self.assertEqual(parser.unparsed, b'remaining-data')

    def test_parse_parsable(self):
        parser = ParserBinary(b'\x01\x02\x03\x04')

        parser.parse_parsable('first_byte', OneByteParsable)
        self.assertEqual(
            b'\x01',
            parser['first_byte'].compose()
        )

        parser.parse_parsable('second_byte', OneByteParsable)
        self.assertEqual(
            b'\x02',
            parser['second_byte'].compose()
        )

        parser = ParserBinary(b'\x01\x02')
        parser.parse_parsable('byte', OneByteParsable, 1)
        self.assertEqual(
            b'\x02',
            parser['byte'].compose()
        )

        parser = ParserBinary(b'\x02\x01\x02')
        with self.assertRaises(TooMuchData) as context_manager:
            parser.parse_parsable('byte', OneByteParsable, 1)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_parse_parsable_array(self):
        parser = ParserBinary(b'\x01\x02\x03\x04')
        parser.parse_parsable_array('array', items_size=4, item_class=OneByteParsable)
        self.assertEqual(
            [0x01, 0x02, 0x03, 0x04],
            list(map(int, parser['array']))
        )

        parser = ParserBinary(b'\x01\x02\x03\x04')
        parser.parse_parsable_array('array', items_size=4, item_class=TwoByteParsable)
        self.assertEqual(
            [0x0102, 0x0304],
            list(map(int, parser['array']))
        )

        parser = ParserBinary(b'\x01\x02')
        with self.assertRaises(InvalidValue):
            parser.parse_parsable_array('array', items_size=2, item_class=OneByteOddParsable)

        parser = ParserBinary(b'\x00')
        with self.assertRaises(NotEnoughData) as context_manager:
            parser.parse_parsable_array('array', items_size=3, item_class=OneByteOddParsable)
        self.assertEqual(context_manager.exception.bytes_needed, 2)

    def test_parse_parsable_derived_array(self):
        parser = ParserBinary(b'\x01\x02\x00')
        parser.parse_parsable_derived_array(
            'array',
            items_size=3,
            item_base_class=ConditionalParsable,
            fallback_class=None
        )
        self.assertEqual(
            [0x01, 0x0200],
            list(map(int, parser['array']))
        )
        self.assertEqual(parser.unparsed_length, 0)
        self.assertEqual(parser.unparsed, b'')

        parser = ParserBinary(b'\x00\x01')
        with self.assertRaises(InvalidValue):
            parser.parse_parsable_derived_array(
                'array',
                items_size=2,
                item_base_class=ConditionalParsable,
                fallback_class=None
            )

        parser = ParserBinary(b'\x00\x01')
        parser.parse_parsable_derived_array(
            'array',
            items_size=2,
            item_base_class=ConditionalParsable,
            fallback_class=TwoByteParsable
        )
        self.assertEqual(
            [0x01, ],
            list(map(int, parser['array']))
        )
        self.assertEqual(parser.unparsed_length, 0)
        self.assertEqual(parser.unparsed, b'')

    def test_parse_variant_parsable(self):
        AlwaysInvalidTypeVariantParsable.register_variant_parser(SerializableEnumFactory, SerializableEnumFactory)
        self.assertEqual(
            AlwaysInvalidTypeVariantParsable.parse_exact_size(b'\x00\x01').value,
            AlwaysInvalidTypeVariantParsable(SerializableEnum.FIRST).variant.value
        )

    def test_parse_timestamp(self):
        parser = ParserBinary(b'\x00\x00\x00\x00\x00\x00\x00\x00')
        parser.parse_timestamp('timestamp')
        self.assertEqual(parser['timestamp'], datetime.datetime.fromtimestamp(0, datetime.timezone.utc))

        parser = ParserBinary(b'\x00\x00\x00\x00')
        parser.parse_timestamp('timestamp', item_size=4)
        self.assertEqual(parser['timestamp'], datetime.datetime.fromtimestamp(0, datetime.timezone.utc))

        parser = ParserBinary(b'\x00\x00\x00\x00\x00\x00\x00\xff')
        parser.parse_timestamp('timestamp', milliseconds=True)
        self.assertEqual(
            parser['timestamp'],
            datetime.datetime.fromtimestamp(0, datetime.timezone.utc) + datetime.timedelta(microseconds=255000)
        )

        parser = ParserBinary(b'\x00\x00\x00\xff')
        parser.parse_timestamp('timestamp', milliseconds=True, item_size=4)
        self.assertEqual(
            parser['timestamp'],
            datetime.datetime.fromtimestamp(0, datetime.timezone.utc) + datetime.timedelta(microseconds=255000)
        )

        parser = ParserBinary(b'\xff\xff\xff\xff\xff\xff\xff\xff')
        parser.parse_timestamp('timestamp')
        self.assertEqual(parser['timestamp'], None)

        parser = ParserBinary(b'\xff\xff\xff\xff')
        parser.parse_timestamp('timestamp', item_size=4)
        self.assertEqual(parser['timestamp'], None)

        parser = ParserBinary(b'\x00\x00\x00\x00\xff\xff\xff\xff')
        parser.parse_timestamp('timestamp')
        self.assertEqual(parser['timestamp'], datetime.datetime.fromtimestamp(0xffffffff, datetime.timezone.utc))

        parser = ParserBinary(b'\x00\x00\xff\xff')
        parser.parse_timestamp('timestamp', item_size=4)
        self.assertEqual(parser['timestamp'], datetime.datetime.fromtimestamp(0x0000ffff, datetime.timezone.utc))


class TestParserText(TestParsableBase):
    def test_error(self):
        parser = ParserText(b'\xff')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_until_separator('string', '#')
        self.assertEqual(context_manager.exception.value, b'\xff')

    def test_separator(self):
        parser = ParserText(b';')
        self.assertEqual(parser.unparsed_length, 1)

        parser.parse_separator(';')
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b';;test;;')
        parser.parse_separator(';', max_length=None)
        self.assertEqual(parser.unparsed_length, 6)

        parser.parse_string_by_length('test', 4, 4)
        self.assertEqual(parser.unparsed_length, 2)

        parser.parse_separator(';', max_length=None)
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b';;')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_separator(';', max_length=1)
        self.assertEqual(context_manager.exception.value, b';;')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_separator(';', min_length=3)
        self.assertEqual(context_manager.exception.value, b';;')

    def test_parse_numeric(self):
        parser = ParserText(b'1#')
        parser.parse_numeric('number')
        self.assertEqual(parser['number'], 1)

        parser = ParserText(b'NaN')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_numeric('number')
        self.assertEqual(context_manager.exception.value, b'NaN')

        parser = ParserText(b'1a')
        parser.parse_numeric('number')
        self.assertEqual(parser['number'], 1)
        self.assertEqual(parser.unparsed_length, 1)

        parser = ParserText(b'1.2#')
        parser.parse_numeric_array('number', 2, '.')
        self.assertEqual(parser['number'], [1, 2])

        parser = ParserText(b'1#')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_numeric_array('number', 2, '.')
        self.assertEqual(context_manager.exception.value, b'1')

    def test_parse_float(self):
        parser = ParserText(b'1.2#')
        parser.parse_float('number')
        self.assertEqual(parser['number'], 1.2)

        parser = ParserText(b'1.#')
        parser.parse_float('number')
        self.assertEqual(parser['number'], 1.0)

        parser = ParserText(b'1#')
        parser.parse_float('number')
        self.assertEqual(parser['number'], 1.0)

        parser = ParserText(b'NaN')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_float('number')
        self.assertEqual(context_manager.exception.value, b'NaN')

    def test_parse_string_until_separator(self):
        parser = ParserText(b'a#')
        parser.parse_string_until_separator('string', '#')
        self.assertEqual(parser['string'], 'a')
        self.assertEqual(parser.unparsed_length, 1)

        parser = ParserText(b'12#')
        parser.parse_string_until_separator('number', '#', int)
        self.assertEqual(parser['number'], 12)
        self.assertEqual(parser.unparsed_length, 1)

        parser = ParserText(b'three#one')
        parser.parse_string_until_separator('number', '#', StringEnum)
        self.assertEqual(parser['number'], StringEnum.THREE)
        self.assertEqual(parser.unparsed_length, 4)

        parser = ParserText(self._ALPHA_BETA_GAMMA_HASHMARK_BYTES, 'utf-8')
        parser.parse_string_until_separator('alphabet', '#')
        self.assertEqual(parser['alphabet'], 'αβγ')
        self.assertEqual(parser.unparsed_length, 1)

        parser = ParserText(b'ab')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_until_separator('string', '#')
        self.assertEqual(context_manager.exception.value, b'ab')

        parser = ParserText(b'ab')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_until_separator('string', '#')
        self.assertEqual(context_manager.exception.value, b'ab')

        parser = ParserText(b'12a#')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_until_separator('string', '#', int)
        self.assertEqual(context_manager.exception.value, b'12a#')

        parser = ParserText(self._ALPHA_BETA_GAMMA_HASHMARK_BYTES, 'ascii')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_until_separator('alphabet', '#')
        self.assertEqual(context_manager.exception.value, self._ALPHA_BETA_GAMMA_HASHMARK_BYTES)

    def test_parse_string_until_separator_or_end(self):
        parser = ParserText(b'ab')
        parser.parse_string_until_separator_or_end('string', '#')
        self.assertEqual(parser['string'], 'ab')
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'12')
        parser.parse_string_until_separator_or_end('number', '#', int)
        self.assertEqual(parser['number'], 12)
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'a#b')
        parser.parse_string_until_separator_or_end('string', '#')
        self.assertEqual(parser['string'], 'a')
        self.assertEqual(parser.unparsed_length, 2)

    def test_parse_string(self):
        parser = ParserText(b'abc')
        parser.parse_string('string', 'abc')
        self.assertEqual(parser['string'], 'abc')

        parser = ParserText(b'abcd')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string('string', 'bcd')
        self.assertEqual(context_manager.exception.value, b'abc')

        parser = ParserText(b'abc')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string('string', 'abcd')
        self.assertEqual(context_manager.exception.value, b'abc')

    def test_parse_string_by_length(self):
        parser = ParserText(b'abc')
        parser.parse_string_by_length('string', 1, None)
        self.assertEqual(parser['string'], 'abc')
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'12')
        parser.parse_string_by_length('number', 1, None, int)
        self.assertEqual(parser['number'], 12)
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'12ab')
        parser.parse_string_by_length('string', 1, 2, int)
        self.assertEqual(parser['string'], 12)
        self.assertEqual(parser.unparsed_length, 2)

        parser = ParserText(b'12')
        with self.assertRaises(NotEnoughData) as context_manager:
            parser.parse_string_by_length('string', 3, 3)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        parser = ParserText(b'12ab')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_by_length('string', 3, 4, int)
        self.assertEqual(context_manager.exception.value, '12ab')

        parser = ParserText(self._ALPHA_BETA_GAMMA_HASHMARK_BYTES, 'ascii')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_by_length('alphabet')
        self.assertEqual(context_manager.exception.value, self._ALPHA_BETA_GAMMA_HASHMARK_BYTES)

    def test_parse_bool(self):
        parser = ParserText(b'yes')
        parser.parse_bool('bool')
        self.assertEqual(parser['bool'], True)
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'no')
        parser.parse_bool('bool')
        self.assertEqual(parser['bool'], False)
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'abcd')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_bool('bool')
        self.assertEqual(context_manager.exception.value, b'abcd')


class TestParserTextStringArray(TestParsableBase):
    def test_empty(self):
        parser = ParserText(b'')
        parser.parse_string_array('array', ',', skip_empty=True)
        self.assertEqual(parser['array'], [])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_array('array', ',', skip_empty=False)
        self.assertEqual(context_manager.exception.value, b'')
        self.assertEqual(parser.unparsed_length, 0)

    def test_separator_only(self):
        parser = ParserText(b',,,')
        parser.parse_string_array('array', ',', skip_empty=True)
        self.assertEqual(parser['array'], [])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b',,,')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_array('array', ',', skip_empty=False)
        self.assertEqual(context_manager.exception.value, b',,,')
        self.assertEqual(parser.unparsed_length, 3)

    def test_space_only(self):
        parser = ParserText(b'  ')
        parser.parse_string_array('array', ',', separator_spaces=' ', skip_empty=True)
        self.assertEqual(parser['array'], [])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'   ')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_array('array', ',', separator_spaces=' ')
        self.assertEqual(context_manager.exception.value, b'')
        self.assertEqual(parser.unparsed_length, 3)

    def test_separator_and_spaces(self):
        parser = ParserText(b'  ,  ,,  ,,,')
        parser.parse_string_array('array', ',', separator_spaces=' ', skip_empty=True)
        self.assertEqual(parser['array'], [])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'  ,  ,,  ,,,')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_array('array', ',', separator_spaces=' ')
        self.assertEqual(context_manager.exception.value, b',  ,,  ,,,')
        self.assertEqual(parser.unparsed_length, 12)

    def test_one_character_separator(self):
        parser = ParserText(b'a,b')
        parser.parse_string_array('array', ',')
        self.assertEqual(parser['array'], ['a', 'b'])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'a,b')
        parser.parse_string_array('array', ',', item_class=ord)
        self.assertEqual(parser['array'], [ord('a'), ord('b')])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'one,two')
        parser.parse_string_array('array', ',', item_class=StringEnum)
        self.assertEqual(parser['array'], [StringEnum.ONE, StringEnum.TWO])
        self.assertEqual(parser.unparsed_length, 0)

    def test_separator_spaces(self):
        parser = ParserText(b' a; \tb\t;\t c')
        parser.parse_string_array('array', ';', separator_spaces=' \t')
        self.assertEqual(parser['array'], ['a', 'b', 'c'])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b' one; \ttwo\t;\t three')
        parser.parse_string_array('array', ';', item_class=StringEnum, separator_spaces=' \t')
        self.assertEqual(parser['array'], [StringEnum.ONE, StringEnum.TWO, StringEnum.THREE])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b' a \t b ; \tc')
        parser.parse_string_array('array', ';', separator_spaces='\t')
        self.assertEqual(parser['array'], [' a \t b ', ' \tc'])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b' a ')
        parser.parse_string_array('array', ';', separator_spaces=' ')
        self.assertEqual(parser['array'], ['a', ])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b' one ')
        parser.parse_string_array('array', ';', item_class=StringEnum, separator_spaces=' ')
        self.assertEqual(parser['array'], [StringEnum.ONE, ])
        self.assertEqual(parser.unparsed_length, 0)

    def test_starts_with_separator(self):
        parser = ParserText(b'; a; b; c')
        parser.parse_string_array('array', ';', separator_spaces=' ', skip_empty=True)
        self.assertEqual(parser['array'], ['a', 'b', 'c'])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'; a; b; c')
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_string_array('array', ';', separator_spaces=' ')
        self.assertEqual(context_manager.exception.value, b'; a; b; c')
        self.assertEqual(parser.unparsed_length, 9)

    def test_ends_with_separator(self):
        parser = ParserText(b'a; b; c; ')
        parser.parse_string_array('array', ';', separator_spaces=' ', skip_empty=True)
        self.assertEqual(parser['array'], ['a', 'b', 'c'])
        self.assertEqual(parser.unparsed_length, 0)

        parser = ParserText(b'one; two; three; ')
        parser.parse_string_array('array', ';', item_class=StringEnum, separator_spaces=' ', skip_empty=True)
        self.assertEqual(parser['array'], [StringEnum.ONE, StringEnum.TWO, StringEnum.THREE])
        self.assertEqual(parser.unparsed_length, 0)

    def test_ends_without_separator(self):
        parser = ParserText(b'a; b; c')
        parser.parse_string_array('array', ';', separator_spaces=' ', max_item_num=2)
        self.assertEqual(parser['array'], ['a', 'b'])
        self.assertEqual(parser.unparsed_length, 1)

        parser = ParserText(b'one; two; three')
        parser.parse_string_array('array', ';', item_class=StringEnum, separator_spaces=' ', max_item_num=2)
        self.assertEqual(parser['array'], [StringEnum.ONE, StringEnum.TWO])
        self.assertEqual(parser.unparsed_length, 5)


class TestParserTextDateTime(TestParsableBase):
    def test_parse_date_time(self):
        parser = ParserText(b'Wed, 21 Oct 2015 07:28:00 GMT')
        parser.parse_date_time('datetime')
        self.assertEqual(
            parser['datetime'],
            datetime.datetime(2015, 10, 21, 7, 28, tzinfo=datetime.timezone.utc)
        )
        self.assertEqual(parser.unparsed_length, 0)

        datetime_value = b'not a date'
        parser = ParserText(datetime_value)
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_date_time('datetime')
        self.assertEqual(context_manager.exception.value, datetime_value)
        self.assertEqual(parser.unparsed_length, len(datetime_value))


class TestParserTextTimeDelta(TestParsableBase):
    def test_parse_time_delta(self):
        parser = ParserText(b'86400')
        parser.parse_time_delta('timedelta')
        self.assertEqual(parser['timedelta'], datetime.timedelta(1))
        self.assertEqual(parser.unparsed_length, 0)

        timedelta_value = str(int(datetime.timedelta.max.total_seconds())).encode('ascii')
        parser = ParserText(timedelta_value)
        with self.assertRaises(InvalidValue) as context_manager:
            parser.parse_time_delta('timedelta')
        self.assertEqual(
            context_manager.exception.value,
            int(datetime.timedelta.max.total_seconds())
        )
        self.assertEqual(parser.unparsed_length, len(timedelta_value))


class TestComposerBinary(TestParsableBase):
    def test_error(self):
        composer = ComposerBinary()

        for size in (1, 2, 4):
            min_value = 0
            max_value = 2 ** (size * 8)

            with self.assertRaises(InvalidValue) as context_manager:
                composer.compose_numeric(max_value + 1, size)
            self.assertEqual(context_manager.exception.value, max_value + 1)

            with self.assertRaises(InvalidValue) as context_manager:
                composer.compose_numeric(min_value - 1, size)
            self.assertEqual(context_manager.exception.value, min_value - 1)

    def test_compose_numeric_to_right_size(self):
        composer = ComposerBinary()
        composer.compose_numeric(0x01, 1)
        self.assertEqual(composer.composed_bytes, b'\x01')

        composer = ComposerBinary()
        composer.compose_numeric(0x01, 2)
        self.assertEqual(composer.composed_bytes, b'\x00\x01')

        composer = ComposerBinary()
        composer.compose_numeric(0x01, 3)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x01')

        composer = ComposerBinary()
        composer.compose_numeric(0x01, 4)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x01')

    def test_compose_numeric_to_rigth_order(self):
        composer = ComposerBinary()
        composer.compose_numeric(0x01, 1)
        self.assertEqual(composer.composed_bytes, b'\x01')

        composer = ComposerBinary()
        composer.compose_numeric(0x0102, 2)
        self.assertEqual(composer.composed_bytes, b'\x01\x02')

        composer = ComposerBinary()
        composer.compose_numeric(0x010203, 3)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03')

        composer = ComposerBinary()
        composer.compose_numeric(0x01020304, 4)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04')

    def test_compose_byte_order(self):
        composer = ComposerBinary(byte_order=ByteOrder.BIG_ENDIAN)
        composer.compose_numeric(0x01020304, 4)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04')

        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)
        composer.compose_numeric(0x01020304, 4)
        self.assertEqual(composer.composed_bytes, b'\x04\x03\x02\x01')

        composer = ComposerBinary(byte_order=ByteOrder.NETWORK)
        composer.compose_numeric(0x01020304, 4)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04')

        composer = ComposerBinary(byte_order=ByteOrder.BIG_ENDIAN)
        composer.compose_numeric(0x010203, 3)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03')

        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)
        composer.compose_numeric(0x010203, 3)
        self.assertEqual(composer.composed_bytes, b'\x03\x02\x01')

        composer = ComposerBinary(byte_order=ByteOrder.NETWORK)
        composer.compose_numeric(0x010203, 3)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03')

    def test_compose_numeric_array(self):
        composer = ComposerBinary()
        composer.compose_numeric_array(values=[1, 2, 3, 4], item_size=1)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04')

        composer = ComposerBinary()
        composer.compose_numeric_array(values=[1, 2, 3, 4], item_size=2)
        self.assertEqual(composer.composed_bytes, b'\x00\x01\x00\x02\x00\x03\x00\x04')

        composer = ComposerBinary()
        composer.compose_numeric_array(values=[1, 2, 3, 4], item_size=3)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x01\x00\x00\x02\x00\x00\x03\x00\x00\x04')

        composer = ComposerBinary()
        composer.compose_numeric_array(values=[1, 2, 3, 4], item_size=4)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04')

    def test_compose_numeric_enum_coded(self):
        composer = ComposerBinary()
        composer.compose_numeric_enum_coded(SerializableEnum.FIRST)
        self.assertEqual(composer.composed_bytes, b'\x00\x01')

    def test_compose_numeric_array_enum_coded(self):
        composer = ComposerBinary()
        composer.compose_numeric_array_enum_coded(values=[])
        self.assertEqual(composer.composed_bytes, b'')

        composer = ComposerBinary()
        composer.compose_numeric_array_enum_coded(values=[
            SerializableEnum.FIRST,
            SerializableEnum.SECOND,
        ])
        self.assertEqual(composer.composed_bytes, b'\x00\x01\x00\x02')

    def test_compose_numeric_flags(self):
        composer = ComposerBinary()
        composer.compose_numeric_flags([FlagEnum.ONE, ], 1)
        self.assertEqual(composer.composed_bytes, b'\x01')

        composer = ComposerBinary()
        composer.compose_numeric_flags([FlagEnum.ONE, FlagEnum.TWO, ], 1)
        self.assertEqual(composer.composed_bytes, b'\x03')

    def test_compose_mpint(self):
        composer = ComposerBinary()
        with self.assertRaises(InvalidValue) as context_manager:
            composer.compose_mpint(1024, 1)
        self.assertEqual(context_manager.exception.value, 1)

        composer = ComposerBinary()
        composer.compose_mpint(1024, 2)
        self.assertEqual(composer.composed_bytes, b'\x04\x00')

        composer = ComposerBinary()
        composer.compose_mpint(1024, 10)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00')

        composer = ComposerBinary()
        composer.compose_mpint(-1024, 10)
        self.assertEqual(composer.composed_bytes, b'\xff\xff\xff\xff\xff\xff\xff\xff\xfc\x00')

        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)
        composer.compose_mpint(1000, 2)
        self.assertEqual(composer.composed_bytes, b'\xe8\x03')

        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)
        composer.compose_mpint(1000, 10)
        self.assertEqual(composer.composed_bytes, b'\xe8\x03\x00\x00\x00\x00\x00\x00\x00\x00')

        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)
        composer.compose_mpint(-1000, 10)
        self.assertEqual(composer.composed_bytes, b'\x18\xfc\xff\xff\xff\xff\xff\xff\xff\xff')

    def test_compose_ssh_mpint(self):
        composer = ComposerBinary()
        composer.compose_ssh_mpint(0)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x00')

        composer = ComposerBinary()
        composer.compose_ssh_mpint(0x9a378f9b2e332a7)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x08\x09\xa3\x78\xf9\xb2\xe3\x32\xa7')

        composer = ComposerBinary()
        composer.compose_ssh_mpint(0x80)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x02\x00\x80')

        composer = ComposerBinary()
        composer.compose_ssh_mpint(-0x1234)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x02\xed\xcc')

        composer = ComposerBinary()
        composer.compose_ssh_mpint(-0xdeadbeef)
        self.assertEqual(composer.composed_bytes, b'\x00\x00\x00\x05\xff\x21\x52\x41\x11')

    def test_compose_raw(self):
        composer = ComposerBinary()

        composer.compose_raw(b'\x01\x02\x03\x04')

        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04')

    def test_compose_string(self):
        composer = ComposerBinary()
        with self.assertRaises(InvalidValue) as context_manager:
            composer.compose_string(self._ALPHA_BETA_GAMMA, 'ascii', 1)
        self.assertEqual(context_manager.exception.value, self._ALPHA_BETA_GAMMA)

        composer = ComposerBinary()
        composer.compose_string(self._ALPHA_BETA_GAMMA, 'utf-8', 1)
        self.assertEqual(composer.composed_bytes[1:], self._ALPHA_BETA_GAMMA_BYTES)

    def test_compose_string_null_terminated(self):
        composer = ComposerBinary()
        with self.assertRaises(InvalidValue) as context_manager:
            composer.compose_string_null_terminated(self._ALPHA_BETA_GAMMA, 'ascii')
        self.assertEqual(context_manager.exception.value, self._ALPHA_BETA_GAMMA)

        composer = ComposerBinary()
        composer.compose_string_null_terminated(self._ALPHA_BETA_GAMMA, 'utf-8')
        self.assertEqual(composer.composed_bytes, self._ALPHA_BETA_GAMMA_BYTES + b'\x00')

    def test_compose_string_enum_coded(self):
        composer = ComposerBinary()
        composer.compose_string_enum_coded(StringEnum.ONE, 2)
        self.assertEqual(composer.composed_bytes, b'\x00\x03one')

    def test_compose_multiple(self):
        composer = ComposerBinary()

        one_byte_parsable = OneByteParsable(0x01)
        composer.compose_parsable(one_byte_parsable)
        self.assertEqual(composer.composed_bytes, b'\x01')

        composer.compose_numeric(0x02, 1)
        self.assertEqual(composer.composed_bytes, b'\x01\x02')
        self.assertEqual(composer.composed_length, 2)

        composer.compose_numeric(0x0304, 2)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04')
        self.assertEqual(composer.composed_length, 4)

        composer.compose_numeric(0x050607, 3)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04\x05\x06\x07')
        self.assertEqual(composer.composed_length, 7)

        composer.compose_numeric(0x08090a0b, 4)
        self.assertEqual(composer.composed_bytes, b'\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b')
        self.assertEqual(composer.composed_length, 11)

    def test_compose_parsable(self):
        composer = ComposerText()
        composer.compose_parsable(StringEnum.THREE)
        composer.compose_parsable(StringEnum.ONE)
        self.assertEqual(composer.composed, b'threeone')

        composer = ComposerBinary()
        composer.compose_parsable(OneByteParsable(0x01))
        composer.compose_parsable(TwoByteParsable(0x0203))
        self.assertEqual(
            b'\x01\x02\x03',
            composer.composed_bytes
        )

    def test_compose_parsable_array(self):
        composer = ComposerText()
        with self.assertRaises(InvalidType):
            composer.compose_parsable_array([
                StringEnum.THREE,
                StringEnum.ONE,
                'four'
            ])

        composer = ComposerText()
        with self.assertRaises(InvalidType):
            composer.compose_parsable_array([
                StringEnum.THREE,
                StringEnum.ONE,
                'four'
            ], fallback_class=int)

        composer = ComposerText()
        composer.compose_parsable_array([
            StringEnum.THREE,
            StringEnum.ONE,
            'four'
        ], fallback_class=str)
        self.assertEqual(composer.composed, b'three,one,four')

        composer = ComposerBinary()
        parsable_array = [OneByteParsable(0x01), TwoByteParsable(0x0203), ]
        composer.compose_parsable_array(parsable_array)

        self.assertEqual(
            b'\x01\x02\x03',
            composer.composed_bytes
        )

    def test_compose_enum(self):
        composer = ComposerBinary()
        composer.compose_parsable(SerializableEnum.SECOND)
        self.assertEqual(b'\x00\x02', composer.composed_bytes)

        composer = ComposerBinary()
        composer.compose_parsable(SerializableEnum.FIRST, item_size=1)
        self.assertEqual(b'\x02\x00\x01', composer.composed_bytes)

    def test_compose_variant_parsable(self):
        composer = ComposerBinary()
        composer.compose_parsable(SerializableEnumVariantParsable(SerializableEnum.FIRST))
        self.assertEqual(b'\x00\x01', composer.composed_bytes)

    def test_compose_timestamp(self):
        composer = ComposerBinary()
        date_time = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)
        composer.compose_timestamp(date_time)
        self.assertEqual(b'\x00\x00\x00\x00\x00\x00\x00\x00', composer.composed_bytes)

        composer = ComposerBinary()
        date_time = datetime.datetime.fromtimestamp(0, datetime.timezone.utc) + datetime.timedelta(microseconds=255000)
        composer.compose_timestamp(date_time, milliseconds=True)
        self.assertEqual(b'\x00\x00\x00\x00\x00\x00\x00\xff', composer.composed_bytes)

        composer = ComposerBinary()
        date_time = datetime.datetime.fromtimestamp(0xffffffff, datetime.timezone.utc)
        date_time.replace(tzinfo=None)
        composer.compose_timestamp(date_time)
        self.assertEqual(b'\x00\x00\x00\x00\xff\xff\xff\xff', composer.composed_bytes)

        composer = ComposerBinary()
        composer.compose_timestamp(None)
        self.assertEqual(b'\xff\xff\xff\xff\xff\xff\xff\xff', composer.composed_bytes)


class TestComposerText(TestParsableBase):
    def test_compose_numeric(self):
        composer = ComposerText()

        composer.compose_numeric(1)
        self.assertEqual(composer.composed, b'1')

        composer.compose_numeric(2)
        self.assertEqual(composer.composed, b'12')
        self.assertEqual(composer.composed_length, 2)

    def test_compose_numeric_array(self):
        composer = ComposerText()

        composer.compose_numeric_array([1, 2], separator=',')
        self.assertEqual(composer.composed, b'1,2')
        self.assertEqual(composer.composed_length, 3)

    def test_compose_string(self):
        composer = ComposerText()

        composer.compose_string('abc')
        self.assertEqual(composer.composed, b'abc')
        self.assertEqual(composer.composed_length, 3)

        composer = ComposerText('utf-8')
        for index, char in enumerate(self._ALPHA_BETA_GAMMA):
            composer.compose_string(char)
            expected_composed = self._ALPHA_BETA_GAMMA[0:index + 1].encode('utf-8')
            self.assertEqual(composer.composed, expected_composed)
            self.assertEqual(composer.composed_length, (index + 1) * 2)

        composer = ComposerText()
        with self.assertRaises(InvalidValue) as context_manager:
            composer.compose_string(self._ALPHA_BETA_GAMMA)
        self.assertEqual(context_manager.exception.value, self._ALPHA_BETA_GAMMA)

    def test_compose_string_array(self):
        composer = ComposerText()

        composer.compose_string_array(['a', 'b', 'c'], '#')
        self.assertEqual(composer.composed, b'a#b#c')
        self.assertEqual(composer.composed_length, 5)

        composer = ComposerText('utf-8')
        composer.compose_string_array(list(self._ALPHA_BETA_GAMMA), '')
        self.assertEqual(composer.composed, self._ALPHA_BETA_GAMMA.encode('utf-8'))
        self.assertEqual(composer.composed_length, len(self._ALPHA_BETA_GAMMA) * 2)

        composer = ComposerText()

        composer.compose_string_array(
            [AlwaysTestStringComposer(), AlwaysTestStringComposer(), AlwaysTestStringComposer()], '#'
        )
        self.assertEqual(composer.composed, b'test#test#test')
        self.assertEqual(composer.composed_length, len(AlwaysTestStringComposer().compose()) * 3 + 2)

    def test_compose_separator(self):
        composer = ComposerText()

        composer.compose_separator('#')
        self.assertEqual(composer.composed, b'#')
        composer.compose_separator('string')
        self.assertEqual(composer.composed, b'#string')
        composer.compose_separator('#')
        self.assertEqual(composer.composed, b'#string#')

        composer = ComposerText('utf-8')

        composer.compose_separator(self._ALPHA_BETA_GAMMA)
        self.assertEqual(composer.composed, self._ALPHA_BETA_GAMMA.encode('utf-8'))

    def test_compose_date_time(self):
        composer = ComposerText()
        composer.compose_date_time(datetime.datetime(2015, 10, 21, 7, 28), '%a, %d %b %Y %H:%M:%S GMT')
        self.assertEqual(composer.composed, b'Wed, 21 Oct 2015 07:28:00 GMT')

    def test_compose_time_delta(self):
        composer = ComposerText()
        composer.compose_time_delta(datetime.timedelta(1))
        self.assertEqual(composer.composed, b'86400')

    def test_compose_bool(self):
        composer = ComposerText()
        composer.compose_bool(True)
        self.assertEqual(composer.composed, b'yes')

        composer = ComposerText()
        composer.compose_bool(False)
        self.assertEqual(composer.composed, b'no')



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import enum
import attr

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary, ByteOrder
from cryptoparser.common.exception import NotEnoughData, InvalidType


@attr.s
class TPKT(ParsableBase):
    HEADER_SIZE = 4

    version = attr.ib(validator=attr.validators.instance_of(int))
    message = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('version', 1)
        if parser['version'] != 3:
            raise InvalidValue(parser['version'], TPKT, 'version')
        parser.parse_numeric('reserved', 1)
        parser.parse_numeric('packet_length', 2)

        if len(parsable) < parser['packet_length']:
            raise NotEnoughData(parser['packet_length'] - len(parsable))

        parser.parse_raw('message', parser['packet_length'] - 4)

        return TPKT(parser['version'], parser['message']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()
        composer.compose_numeric(self.version, 1)
        composer.compose_numeric(0, 1)  # reserved
        composer.compose_numeric(len(self.message) + 4, 2)
        composer.compose_raw(self.message)

        return composer.composed_bytes


class COTPType(enum.IntEnum):
    CONNECTION_REQUEST = 0xe
    CONNECTION_CONFIRM = 0xd
    DISCONNECT_REQUEST = 0x8
    DISCONNECT_CONFIRM = 0xc
    DATA = 0xf
    EXPEDITED_DATA = 0x1
    DATA_ACKNOWLEDGEMENT = 0x6
    EXPEDITED_DATA_ANOWLEDGEMENT = 0x2
    REJECT = 0x5


@attr.s
class COTPConnectionBase(ParsableBase):
    HEADER_SIZE = 7

    src_ref = attr.ib(validator=attr.validators.instance_of(int))
    user_data = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    dst_ref = attr.ib(default=0, validator=attr.validators.instance_of(int))
    class_option = attr.ib(default=0)

    @classmethod
    @abc.abstractmethod
    def _get_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('length_indicator', 1)
        if parser.unparsed_length < parser['length_indicator']:
            raise NotEnoughData(parser['length_indicator'] - parser.unparsed_length)

        parser.parse_numeric('pdu_type', 1)
        pdu_type = parser['pdu_type'] >> 4
        if pdu_type != cls._get_type():
            raise InvalidType()

        parser.parse_numeric('src_ref', 2)
        parser.parse_numeric('dst_ref', 2)
        parser.parse_numeric('class_option', 1)

        parser.parse_raw('user_data', parser['length_indicator'] - parser.parsed_length + 1)

        return COTPConnectionRequest(
            src_ref=parser['src_ref'],
            dst_ref=parser['dst_ref'],
            class_option=parser['class_option'],
            user_data=parser['user_data'],
        ), parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()

        body_composer.compose_numeric(self._get_type() << 4, 1)
        body_composer.compose_numeric(self.src_ref, 2)
        body_composer.compose_numeric(self.dst_ref, 2)
        body_composer.compose_numeric(self.class_option, 1)
        body_composer.compose_raw(self.user_data)

        body = body_composer.composed_bytes

        header_composer = ComposerBinary()
        header_composer.compose_numeric(len(body), 1)

        return header_composer.composed_bytes + body

    def __attrs_post_init__(self):
        if self.class_option != 0:
            raise InvalidValue(self.class_option, COTPConnectionRequest, 'class_option')


@attr.s
class COTPConnectionRequest(COTPConnectionBase):
    @classmethod
    def _get_type(cls):
        return COTPType.CONNECTION_REQUEST


@attr.s
class COTPConnectionConfirm(COTPConnectionBase):
    @classmethod
    def _get_type(cls):
        return COTPType.CONNECTION_CONFIRM


class RDPProtocol(enum.IntEnum):
    RDP = 0x00000000
    SSL = 0x00000001
    HYBRID = 0x00000002
    RDSTLS = 0x00000004
    HYBRID_EX = 0x00000008


class RDPNegotiationRequestFlags(enum.IntEnum):
    RESTRICTED_ADMIN_MODE_REQUIRED = 0x01
    REDIRECTED_AUTHENTICATION_MODE_REQUIRED = 0x02
    CORRELATION_INFO_PRESENT = 0x08


class RDPNegotiationResponseFlags(enum.IntEnum):
    EXTENDED_CLIENT_DATA_SUPPORTED = 0x01
    DYNVC_GFX_PROTOCOL_SUPPORTED = 0x02
    NEGRSP_FLAG_RESERVED = 0x04
    RESTRICTED_ADMIN_MODE_SUPPORTED = 0x08
    REDIRECTED_AUTHENTICATION_MODE_SUPPORTED = 0x10


class RDPPacketType(enum.IntEnum):
    NEG_REQ = 1
    NEG_RSP = 2


@attr.s
class RDPNegotiationBase(ParsableBase):
    PACKET_LENGTH = 8

    flags = attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(
        (RDPNegotiationRequestFlags, RDPNegotiationResponseFlags)
        )))
    protocol = attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.in_(RDPProtocol)))

    @classmethod
    @abc.abstractmethod
    def _get_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_flag_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.PACKET_LENGTH:
            raise NotEnoughData(cls.PACKET_LENGTH - len(parsable))

        parser = ParserBinary(parsable, ByteOrder.LITTLE_ENDIAN)

        parser.parse_numeric('type', 1, RDPPacketType)
        if parser['type'] != cls._get_type():
            raise InvalidType()
        parser.parse_numeric_flags('flags', 1, cls._get_flag_type())
        parser.parse_numeric('length', 2)
        if parser['length'] != cls.PACKET_LENGTH:
            raise InvalidValue(parser['length'], cls, 'packet length')
        parser.parse_numeric_flags('protocol', 4, RDPProtocol)

        return cls(parser['flags'], parser['protocol']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary(ByteOrder.LITTLE_ENDIAN)

        composer.compose_numeric(self._get_type(), 1)
        composer.compose_numeric_flags(self.flags, 1)
        composer.compose_numeric(self.PACKET_LENGTH, 2)
        composer.compose_numeric_flags(self.protocol, 4)

        return composer.composed_bytes


@attr.s
class RDPNegotiationRequest(RDPNegotiationBase):
    @classmethod
    def _get_type(cls):
        return RDPPacketType.NEG_REQ

    @classmethod
    def _get_flag_type(cls):
        return RDPNegotiationRequestFlags


@attr.s
class RDPNegotiationResponse(RDPNegotiationBase):
    @classmethod
    def _get_type(cls):
        return RDPPacketType.NEG_RSP

    @classmethod
    def _get_flag_type(cls):
        return RDPNegotiationResponseFlags



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import collections
import enum

import attr

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.grade import Grade, GradeableSimple

from cryptoparser.common.base import ProtocolVersionBase, Serializable, VariantParsable
from cryptoparser.common.exception import InvalidType
from cryptoparser.common.parse import ParsableBase, ParserText, ComposerText


class SshVersion(enum.IntEnum):
    SSH1 = 1
    SSH2 = 2


@attr.s(hash=True)
class SshProtocolVersion(ProtocolVersionBase, GradeableSimple):
    major = attr.ib(converter=SshVersion, validator=attr.validators.instance_of(SshVersion))
    minor = attr.ib(validator=attr.validators.instance_of(int), default=0)

    @property
    def grade(self):
        if self.major == SshVersion.SSH1:
            return Grade.INSECURE

        return Grade.SECURE

    def __str__(self):
        return f'SSH {self.major}.{self.minor}'

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        try:
            parser.parse_numeric('major')
            parser.parse_separator('.')
            parser.parse_numeric('minor')
        except InvalidValue as e:
            raise InvalidValue(parsable, SshProtocolVersion) from e

        return SshProtocolVersion(parser['major'], parser['minor']), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_numeric(self.major.value)
        composer.compose_separator('.')
        composer.compose_numeric(self.minor)

        return composer.composed

    @property
    def identifier(self):
        return f'ssh{self.major}'

    @property
    def supported_versions(self):
        if self.major == SshVersion.SSH1 and self.minor == 99:
            return [SshVersion.SSH1, SshVersion.SSH2]

        return [self.major, ]


class SshSoftwareVersionBase(ParsableBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class SshSoftwareVersionUnparsed(SshSoftwareVersionBase, Serializable):
    raw = attr.ib(validator=attr.validators.instance_of(str))

    @raw.validator
    def raw_validator(self, _, value):  # pylint: disable=no-self-use
        if '\r' in value or '\n' in value or ' ' in value:
            raise InvalidValue(value, SshSoftwareVersionUnparsed, 'raw')
        try:
            value.encode('ascii')
        except UnicodeEncodeError as e:
            raise InvalidValue(value, SshSoftwareVersionUnparsed, 'raw') from e

    def _as_markdown(self, level):
        return self._markdown_result(self.raw, level)

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_by_length('raw', len(parsable))

        return SshSoftwareVersionUnparsed(parser['raw']), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.raw)

        return composer.composed


@attr.s
class SshSoftwareVersionParsedBase(SshSoftwareVersionBase):
    version = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))

    @classmethod
    @abc.abstractmethod
    def _get_vendor(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_version_separator(cls):
        raise NotImplementedError()

    @property
    def vendor(self):
        return self._get_vendor()

    def _asdict(self):
        return collections.OrderedDict([
            ('vendor', self.vendor),
            ('version', self.version),
        ])

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        version_separator = cls._get_version_separator()
        if version_separator is None:
            parser.parse_string_by_length('vendor')
        else:
            parser.parse_string_until_separator_or_end('vendor', version_separator)

        if parser['vendor'] != cls._get_vendor():
            raise InvalidType()

        if parser.unparsed_length > 0 and version_separator is not None:
            parser.parse_separator(version_separator)
            parser.parse_string_by_length('version')
            version = parser['version']
        else:
            version = None

        return cls(version), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.vendor)
        if self.version is not None:
            composer.compose_separator(self._get_version_separator())
            composer.compose_string(self.version)

        return composer.composed


@attr.s
class SshSoftwareVersionCryptlib(SshSoftwareVersionParsedBase):
    @classmethod
    def _get_vendor(cls):
        return 'cryptlib'

    @classmethod
    def _get_version_separator(cls):
        return None


@attr.s
class SshSoftwareVersionDropbear(SshSoftwareVersionParsedBase):
    @classmethod
    def _get_vendor(cls):
        return 'dropbear'

    @classmethod
    def _get_version_separator(cls):
        return '_'


@attr.s
class SshSoftwareVersionIPSSH(SshSoftwareVersionParsedBase):
    @classmethod
    def _get_vendor(cls):
        return 'IPSSH'

    @classmethod
    def _get_version_separator(cls):
        return '-'


@attr.s
class SshSoftwareVersionMonacaSSH(SshSoftwareVersionParsedBase):
    @classmethod
    def _get_vendor(cls):
        return 'Monaca'

    @classmethod
    def _get_version_separator(cls):
        return None


@attr.s
class SshSoftwareVersionOpenSSH(SshSoftwareVersionParsedBase):
    @classmethod
    def _get_vendor(cls):
        return 'OpenSSH'

    @classmethod
    def _get_version_separator(cls):
        return '_'


class SshSoftwareVersionParsedVariant(VariantParsable):
    _VARIANTS = collections.OrderedDict([
        (SshSoftwareVersionCryptlib, (SshSoftwareVersionCryptlib, )),
        (SshSoftwareVersionDropbear, (SshSoftwareVersionDropbear, )),
        (SshSoftwareVersionIPSSH, (SshSoftwareVersionIPSSH, )),
        (SshSoftwareVersionMonacaSSH, (SshSoftwareVersionMonacaSSH, )),
        (SshSoftwareVersionOpenSSH, (SshSoftwareVersionOpenSSH, )),
    ])

    @classmethod
    def _get_variants(cls):
        return cls._VARIANTS



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import collections
import datetime
import enum

import attr

import asn1crypto

from cryptodatahub.common.key import PublicKeyX509Base
from cryptodatahub.common.stores import CertificateTransparencyLog, CertificateTransparencyLogParamsBase

from cryptoparser.common.base import (
    Opaque,
    OpaqueParam,
    Serializable,
    VectorParamParsable,
    VectorParsable,
)
from cryptoparser.common.parse import ComposerBinary, ParsableBase, ParserBinary

from cryptoparser.tls.algorithm import TlsSignatureAndHashAlgorithm, TlsSignatureAndHashAlgorithmFactory


class CtExtensions(Opaque):
    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=0,
            max_byte_num=2 ** 16 - 1,
        )


class CtSignature(Opaque):
    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=0,
            max_byte_num=2 ** 16 - 1,
        )


class CtVersion(enum.IntEnum):
    V1 = 0x00  # pylint: disable=invalid-name


@attr.s
class SignedCertificateTimestamp(ParsableBase, Serializable):
    version = attr.ib(validator=attr.validators.in_(CtVersion))
    log = attr.ib(
        converter=CertificateTransparencyLog.from_log_id,
        validator=attr.validators.instance_of(CertificateTransparencyLogParamsBase)
    )
    timestamp = attr.ib(validator=attr.validators.instance_of(datetime.datetime))
    extensions = attr.ib(
        validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(CtExtensions))
    )
    signature_algorithm = attr.ib(validator=attr.validators.in_(TlsSignatureAndHashAlgorithm))
    signature = attr.ib(
        converter=CtSignature,
        validator=attr.validators.instance_of(CtSignature),
        metadata={'human_friendly': False},
    )

    @classmethod
    def _parse(cls, parsable):
        header_parser = ParserBinary(parsable)
        header_parser.parse_bytes('sct', 2)

        body_parser = ParserBinary(header_parser['sct'])

        body_parser.parse_numeric('version', 1, CtVersion)
        body_parser.parse_raw('log', 32)
        body_parser.parse_timestamp('timestamp', milliseconds=True)
        body_parser.parse_parsable('extensions', CtExtensions)
        body_parser.parse_parsable('signature_algorithm', TlsSignatureAndHashAlgorithmFactory)
        body_parser.parse_parsable('signature', CtSignature)

        return cls(**body_parser), header_parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()

        body_composer.compose_numeric(self.version, 1)
        body_composer.compose_raw(self.log.log_id.value)
        body_composer.compose_timestamp(self.timestamp, milliseconds=True)
        body_composer.compose_parsable(self.extensions)
        body_composer.compose_numeric_enum_coded(self.signature_algorithm)
        body_composer.compose_parsable(self.signature)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(len(body_composer.composed_bytes), 2)

        return header_composer.composed_bytes + body_composer.composed_bytes


class SignedCertificateTimestampList(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=SignedCertificateTimestamp,
            fallback_class=None,
            min_byte_num=0, max_byte_num=2 ** 16 - 1
        )


class PublicKeyX509(PublicKeyX509Base):
    @property
    def signed_certificate_timestamps(self):
        for extension in self._certificate['tbs_certificate']['extensions']:
            if extension['extn_id'].dotted == '1.3.6.1.4.1.11129.2.4.2':
                asn1_value = asn1crypto.core.load(bytes(extension['extn_value']))
                return SignedCertificateTimestampList.parse_exact_size(bytes(asn1_value))

        return SignedCertificateTimestampList([])

    def _asdict(self):
        dict_value = super()._asdict()

        return collections.OrderedDict(list(dict_value.items()) + [
            ('signed_certificate_timestamps', self.signed_certificate_timestamps),
        ])



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import collections
import unittest

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.ike.algorithm import (
    Ikev2DiffieHellmanGroup,
    Ikev2EncryptionAlgorithm,
    Ikev2IntegrityAlgorithm,
    Ikev2ProtocolId,
    Ikev2PseudorandomFunction,
    Ikev2TransformType,
)

from cryptoparser.common.exception import NotEnoughData
from cryptoparser.ike.ikev2 import (
    Ikev2Proposal,
    Ikev2ProposalNextPayload,
    Ikev2PayloadSecurityAssociation,
    Ikev2PayloadFlags,
    Ikev2PayloadType,
    Ikev2TransformPrf,
    Ikev2TransformDhGroup,
    Ikev2TransformEncryptionAlgorithm,
    Ikev2TransformIntegrity,
    TransformAttributeSignatureAlgorithm,
    TransformNextPayload,
)

from .classes import TransformTest


class TestTransformAttributeSignatureAlgorithm(unittest.TestCase):
    def setUp(self):
        self.signature_algorithm = (
            b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
            b'\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
        )

        self.signature_algorithm_payload = TransformAttributeSignatureAlgorithm(
            signature_algorithm=self.signature_algorithm
        )

        self.signature_algorithm_dict = collections.OrderedDict([
            ('format', b'\x00'),  # TYPE_LENGTH_VALUE
            ('type', b'\x12'),  # SIGNATURE_ALGORITHM = 18
            ('length', b'\x00\x20'),  # 32 bytes length
            ('signature_algorithm', self.signature_algorithm),
        ])
        self.signature_algorithm_bytes = b''.join(self.signature_algorithm_dict.values())

    def test_parse(self):
        parsed_signature_algorithm = TransformAttributeSignatureAlgorithm.parse_exact_size(
            self.signature_algorithm_bytes
        )
        self.assertEqual(parsed_signature_algorithm.signature_algorithm, self.signature_algorithm)

    def test_compose(self):
        composed_bytes = self.signature_algorithm_payload.compose()

        parsed_signature_algorithm = TransformAttributeSignatureAlgorithm.parse_exact_size(composed_bytes)
        self.assertEqual(
            parsed_signature_algorithm.signature_algorithm,
            self.signature_algorithm_payload.signature_algorithm
        )

    def test_round_trip(self):
        composed_bytes = self.signature_algorithm_payload.compose()
        parsed_payload: TransformAttributeSignatureAlgorithm = (
            TransformAttributeSignatureAlgorithm.parse_exact_size(composed_bytes)
        )

        self.assertEqual(parsed_payload.signature_algorithm, self.signature_algorithm_payload.signature_algorithm)

    def test_error_parse_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            TransformAttributeSignatureAlgorithm.parse_exact_size(b'\x00\x12')
        self.assertEqual(context_manager.exception.bytes_needed, 2)

        incomplete = self.signature_algorithm_bytes[:-1]
        with self.assertRaises(NotEnoughData) as context_manager:
            TransformAttributeSignatureAlgorithm.parse_exact_size(incomplete)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_error_invalid_type(self):
        wrong_type_header = b'\x00\x11\x00\x20'  # wrong type 0x11 instead of 0x12
        wrong_type_bytes = wrong_type_header + self.signature_algorithm
        with self.assertRaises(Exception):
            TransformAttributeSignatureAlgorithm.parse_exact_size(wrong_type_bytes)

    def test_error_invalid_signature_algorithm_type(self):
        with self.assertRaises(TypeError):
            TransformAttributeSignatureAlgorithm(
                signature_algorithm="not_bytes"
            )


class TestTransform(unittest.TestCase):
    def setUp(self):
        self.transform_id = Ikev2PseudorandomFunction.PRF_HMAC_SHA1

        self.transform = TransformTest(
            transform_id=self.transform_id
        )
        self.transform.next_payload = TransformNextPayload.LAST

        self.transform_dict = collections.OrderedDict([
            ('next_payload', b'\x00'),  # LAST
            ('reserved1', b'\x00'),
            ('transform_length', b'\x00\x08'),  # 8 bytes header only
            ('transform_type', b'\x02'),  # PRF
            ('reserved2', b'\x00'),
            ('transform_id', b'\x00\x02'),  # PRF_HMAC_SHA1
        ])
        self.transform_bytes = b''.join(self.transform_dict.values())

    def test_parse(self):
        parsed_transform = TransformTest.parse_exact_size(self.transform_bytes)
        self.assertEqual(parsed_transform.transform_id, self.transform_id)
        self.assertEqual(parsed_transform.next_payload, TransformNextPayload.LAST)

    def test_compose(self):
        composed_bytes = self.transform.compose()
        self.assertEqual(len(composed_bytes), 8)  # Header only, no attributes

        parsed_transform = TransformTest.parse_exact_size(composed_bytes)
        self.assertEqual(parsed_transform.transform_id, self.transform.transform_id)

    def test_round_trip(self):
        composed_bytes = self.transform.compose()
        parsed_payload: TransformTest = TransformTest.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_payload.transform_id, self.transform.transform_id)
        self.assertEqual(parsed_payload.next_payload, self.transform.next_payload)

    def test_get_transform_type(self):
        self.assertEqual(TransformTest.get_transform_type(), Ikev2TransformType.PRF)

    def test_get_transform_id_class(self):
        self.assertEqual(
            TransformTest._get_transform_id_class(),  # pylint: disable=protected-access
            Ikev2PseudorandomFunction
        )

    def test_error_parse_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            TransformTest.parse_exact_size(b'\x00\x01\x02')
        self.assertEqual(context_manager.exception.bytes_needed, 5)

        incomplete = self.transform_bytes[:-1]
        with self.assertRaises(NotEnoughData) as context_manager:
            TransformTest.parse_exact_size(incomplete)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_error_invalid_transform_id(self):
        with self.assertRaises(InvalidValue):
            TransformTest(
                transform_id="invalid_id"
            )


class TestIkev2TransformPrf(unittest.TestCase):
    def test_get_transform_type(self):
        self.assertEqual(Ikev2TransformPrf.get_transform_type(), Ikev2TransformType.PRF)

    def test_get_transform_id_class(self):
        self.assertEqual(
            Ikev2TransformPrf._get_transform_id_class(),  # pylint: disable=protected-access
            Ikev2PseudorandomFunction
        )


class TestIkev2TransformDhGroup(unittest.TestCase):
    def test_get_transform_type(self):
        self.assertEqual(Ikev2TransformDhGroup.get_transform_type(), Ikev2TransformType.DH)

    def test_get_transform_id_class(self):
        self.assertEqual(
            Ikev2TransformDhGroup._get_transform_id_class(),  # pylint: disable=protected-access
            Ikev2DiffieHellmanGroup
        )


class TestIkev2TransformIntegrity(unittest.TestCase):
    def test_get_transform_type(self):
        self.assertEqual(Ikev2TransformIntegrity.get_transform_type(), Ikev2TransformType.INTEG)

    def test_get_transform_id_class(self):
        self.assertEqual(
            Ikev2TransformIntegrity._get_transform_id_class(),  # pylint: disable=protected-access
            Ikev2IntegrityAlgorithm
        )


class TestIkev2TransformEncryptionAlgorithm(unittest.TestCase):
    def setUp(self):
        self.encryption_algorithm = Ikev2EncryptionAlgorithm.ENCR_AES_CBC
        self.key_length = 128
        self.transform = Ikev2TransformEncryptionAlgorithm(
            transform_id=self.encryption_algorithm,
            key_length=self.key_length
        )
        self.transform.next_payload = TransformNextPayload.LAST

    def test_get_transform_type(self):
        self.assertEqual(Ikev2TransformEncryptionAlgorithm.get_transform_type(), Ikev2TransformType.ENCR)

    def test_get_transform_id_class(self):
        self.assertEqual(
            Ikev2TransformEncryptionAlgorithm._get_transform_id_class(),  # pylint: disable=protected-access
            Ikev2EncryptionAlgorithm
        )

    def test_key_length_value_support(self):
        different_key_lengths = [128, 192, 256]

        for key_length in different_key_lengths:
            transform = Ikev2TransformEncryptionAlgorithm(
                transform_id=self.encryption_algorithm,
                key_length=key_length
            )
            self.assertEqual(transform.key_length, key_length)

    def test_parse(self):
        composed_bytes = self.transform.compose()
        parsed_transform = Ikev2TransformEncryptionAlgorithm.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_transform.transform_id, self.transform.transform_id)
        self.assertEqual(parsed_transform.key_length, self.transform.key_length)  # pylint: disable=no-member

    def test_compose(self):
        composed_bytes = self.transform.compose()
        self.assertGreater(len(composed_bytes), 8)

    def test_round_trip(self):
        composed_bytes = self.transform.compose()
        parsed_transform = Ikev2TransformEncryptionAlgorithm.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_transform.transform_id, self.transform.transform_id)
        self.assertEqual(parsed_transform.key_length, self.transform.key_length)  # pylint: disable=no-member

    def test_error_parse_not_enough_data(self):
        incomplete_data = b'\x00\x00\x00\x08\x01\x00\x00\x0c'

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev2TransformEncryptionAlgorithm.parse_exact_size(incomplete_data)
        self.assertGreater(context_manager.exception.bytes_needed, 0)


class TestIkev2Proposal(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.transform_prf = Ikev2TransformPrf(
            transform_id=Ikev2PseudorandomFunction.PRF_HMAC_SHA1
        )
        self.transform_prf.next_payload = TransformNextPayload.LAST

        self.protocol_id = Ikev2ProtocolId.IKE
        self.spi = b'\x00\x01\x02\x03'

        self.proposal_minimal = Ikev2Proposal(
            protocol_id=self.protocol_id,
            transforms=[self.transform_prf],
            spi=bytes()
        )
        self.proposal_minimal.last = Ikev2ProposalNextPayload.LAST
        self.proposal_minimal.proposal_number = 1

        self.proposal_with_spi = Ikev2Proposal(
            protocol_id=self.protocol_id,
            transforms=[self.transform_prf],
            spi=self.spi
        )
        self.proposal_with_spi.last = Ikev2ProposalNextPayload.MORE
        self.proposal_with_spi.proposal_number = 2

        self.proposal_dict_minimal = collections.OrderedDict([
            ('last', b'\x00'),  # LAST
            ('reserved', b'\x00'),
            ('proposal_length', b'\x00\x10'),  # 16 bytes total (8 header + 8 transform)
            ('proposal_number', b'\x01'),  # 1
            ('protocol_id', b'\x01'),  # IKE
            ('spi_size', b'\x00'),  # 0 bytes
            ('transform_count', b'\x01'),  # 1 transform
            # No SPI data
            # Transform data: next_payload(1) + reserved(1) + length(2) + type(1) + reserved(1) + id(2)
            ('transform_data', b'\x00\x00\x00\x08\x02\x00\x00\x02'),  # PRF_HMAC_SHA1
        ])
        self.proposal_bytes_minimal = b''.join(self.proposal_dict_minimal.values())
        self.proposal_minimal = Ikev2Proposal(
            protocol_id=self.protocol_id,
            transforms=[self.transform_prf],
            spi=bytes()
        )
        self.proposal_minimal.last = Ikev2ProposalNextPayload.LAST
        self.proposal_minimal.proposal_number = 1

        self.proposal_dict_with_spi = collections.OrderedDict([
            ('last', b'\x02'),  # MORE enum value
            ('reserved', b'\x00'),
            ('proposal_length', b'\x00\x10'),  # 16 bytes total (8 header + 8 transform)
            ('proposal_number', b'\x02'),  # 2
            ('protocol_id', b'\x01'),  # IKE
            ('spi_size', b'\x04'),  # 4 bytes
            ('transform_count', b'\x01'),  # 1 transform
            ('spi', self.spi),  # SPI data
            ('transform_data', b'\x00\x00\x00\x08\x02\x00\x00\x02'),  # PRF_HMAC_SHA1
        ])
        self.proposal_bytes_with_spi = b''.join(self.proposal_dict_with_spi.values())
        self.proposal_with_spi = Ikev2Proposal(
            protocol_id=self.protocol_id,
            transforms=[self.transform_prf],
            spi=self.spi
        )
        self.proposal_with_spi.last = Ikev2ProposalNextPayload.MORE
        self.proposal_with_spi.proposal_number = 2

    def test_parse(self):
        parsed_proposal = Ikev2Proposal.parse_exact_size(self.proposal_bytes_minimal)
        self.assertEqual(parsed_proposal.protocol_id, self.protocol_id)
        self.assertEqual(parsed_proposal.spi, bytes())
        self.assertEqual(len(parsed_proposal.transforms), 1)
        self.assertEqual(parsed_proposal.last, set())
        self.assertEqual(parsed_proposal.proposal_number, 1)

        parsed_proposal_spi = Ikev2Proposal.parse_exact_size(self.proposal_bytes_with_spi)
        self.assertEqual(parsed_proposal_spi.protocol_id, self.protocol_id)
        self.assertEqual(parsed_proposal_spi.spi, self.spi)
        self.assertEqual(len(parsed_proposal_spi.transforms), 1)
        self.assertEqual(parsed_proposal_spi.last, {Ikev2ProposalNextPayload.MORE})
        self.assertEqual(parsed_proposal_spi.proposal_number, 2)

    def test_compose(self):
        composed_bytes = self.proposal_minimal.compose()
        self.assertEqual(composed_bytes, self.proposal_bytes_minimal)

        composed_bytes = self.proposal_with_spi.compose()
        self.assertEqual(composed_bytes, self.proposal_bytes_with_spi)

    def test_round_trip(self):
        composed_bytes = self.proposal_minimal.compose()
        parsed_proposal: Ikev2Proposal = Ikev2Proposal.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_proposal.protocol_id, self.proposal_minimal.protocol_id)
        self.assertEqual(parsed_proposal.spi, self.proposal_minimal.spi)
        self.assertEqual(len(parsed_proposal.transforms), len(self.proposal_minimal.transforms))
        self.assertEqual(parsed_proposal.last, set())
        self.assertEqual(parsed_proposal.proposal_number, self.proposal_minimal.proposal_number)

    def test_error_parse_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev2Proposal.parse_exact_size(b'\x00\x01\x02')
        self.assertEqual(context_manager.exception.bytes_needed, Ikev2Proposal.HEADER_SIZE - 3)

        incomplete = self.proposal_bytes_minimal[:-1]
        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev2Proposal.parse_exact_size(incomplete)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_error_invalid_protocol_id(self):
        with self.assertRaises(TypeError):
            Ikev2Proposal(
                protocol_id="invalid_protocol",
                transforms=[self.transform_prf]
            )

    def test_error_invalid_transforms(self):
        with self.assertRaises(TypeError):
            Ikev2Proposal(
                protocol_id=self.protocol_id,
                transforms=["invalid_transform"]
            )

    def test_error_invalid_spi_type(self):
        with self.assertRaises(TypeError):
            Ikev2Proposal(
                protocol_id=self.protocol_id,
                transforms=[self.transform_prf],
                spi="invalid_spi"
            )


class TestIkev2PayloadSecurityAssociation(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.transform_prf = Ikev2TransformPrf(
            transform_id=Ikev2PseudorandomFunction.PRF_HMAC_SHA1
        )
        self.transform_prf.next_payload = TransformNextPayload.LAST

        self.proposal_single = Ikev2Proposal(
            protocol_id=Ikev2ProtocolId.IKE,
            transforms=[self.transform_prf],
            spi=bytes()
        )

        self.proposal_with_spi = Ikev2Proposal(
            protocol_id=Ikev2ProtocolId.IKE,
            transforms=[self.transform_prf],
            spi=b'\x00\x01\x02\x03'
        )

        self.sa_single_proposal = Ikev2PayloadSecurityAssociation(
            flags=set(),
            proposals=[self.proposal_single]
        )
        self.sa_single_proposal.next_payload = Ikev2PayloadType.NONE

        self.sa_multiple_proposals = Ikev2PayloadSecurityAssociation(
            flags={Ikev2PayloadFlags.CRITICAL},
            proposals=[self.proposal_single, self.proposal_with_spi]
        )
        self.sa_multiple_proposals.next_payload = Ikev2PayloadType.KE

        # Single proposal: header(4) + proposal(16) = 20 bytes total
        self.sa_dict_single = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE
            ('flags', b'\x00'),  # No flags
            ('payload_length', b'\x00\x14'),  # 20 bytes total
            ('proposals_data', b'\x00\x00\x00\x10\x01\x01\x00\x01\x00\x00\x00\x08\x02\x00\x00\x02'),
        ])
        self.sa_bytes_single = b''.join(self.sa_dict_single.values())

        # Multiple proposals: header(4) + proposal1(16) + proposal2(20) = 40 bytes total
        self.sa_dict_multiple = collections.OrderedDict([
            ('next_payload', b'\x22'),  # KE
            ('flags', b'\x80'),  # CRITICAL
            ('payload_length', b'\x00\x28'),  # 40 bytes total
            ('proposals_data', (
                b'\x02\x00\x00\x10\x01\x01\x00\x01\x00\x00\x00\x08\x02\x00\x00\x02' +  # proposal 1
                b'\x00\x00\x00\x10\x02\x01\x04\x01\x00\x01\x02\x03\x00\x00\x00\x08\x02\x00\x00\x02'  # proposal 2
            )),
        ])
        self.sa_bytes_multiple = b''.join(self.sa_dict_multiple.values())

    def test_parse(self):
        parsed_sa = Ikev2PayloadSecurityAssociation.parse_exact_size(self.sa_bytes_single)
        self.assertEqual(parsed_sa.next_payload, Ikev2PayloadType.NONE)
        self.assertEqual(parsed_sa.flags, set())
        self.assertEqual(len(parsed_sa.proposals), 1)
        self.assertEqual(parsed_sa.proposals[0].protocol_id, Ikev2ProtocolId.IKE)

        parsed_sa_multiple = Ikev2PayloadSecurityAssociation.parse_exact_size(self.sa_bytes_multiple)
        self.assertEqual(parsed_sa_multiple.next_payload, Ikev2PayloadType.KE)
        self.assertEqual(parsed_sa_multiple.flags, {Ikev2PayloadFlags.CRITICAL})
        self.assertEqual(len(parsed_sa_multiple.proposals), 2)

    def test_compose(self):
        composed_bytes = self.sa_single_proposal.compose()
        self.assertEqual(composed_bytes, self.sa_bytes_single)

        composed_bytes_multiple = self.sa_multiple_proposals.compose()
        self.assertEqual(composed_bytes_multiple, self.sa_bytes_multiple)

    def test_round_trip(self):
        composed_bytes = self.sa_single_proposal.compose()
        parsed_sa: Ikev2PayloadSecurityAssociation = Ikev2PayloadSecurityAssociation.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_sa.next_payload, self.sa_single_proposal.next_payload)
        self.assertEqual(parsed_sa.flags, self.sa_single_proposal.flags)
        self.assertEqual(len(parsed_sa.proposals), len(self.sa_single_proposal.proposals))

    def test_get_payload_type(self):
        self.assertEqual(Ikev2PayloadSecurityAssociation.get_payload_type(), Ikev2PayloadType.SA)

    def test_error_parse_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev2PayloadSecurityAssociation.parse_exact_size(b'\x00\x01\x02')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        incomplete = self.sa_bytes_single[:-1]
        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev2PayloadSecurityAssociation.parse_exact_size(incomplete)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_error_invalid_proposals(self):
        with self.assertRaises(TypeError):
            Ikev2PayloadSecurityAssociation(
                flags=set(),
                proposals=["invalid_proposal"]
            )

    def test_get_transform_by_type(self):
        transform_prf = Ikev2TransformPrf(
            transform_id=Ikev2PseudorandomFunction.PRF_HMAC_SHA1
        )
        transform_prf.next_payload = TransformNextPayload.LAST

        transform_dh = Ikev2TransformDhGroup(
            transform_id=Ikev2DiffieHellmanGroup.MODP_GROUP_2048_BIT
        )
        transform_dh.next_payload = TransformNextPayload.LAST

        transform_encr = Ikev2TransformEncryptionAlgorithm(
            transform_id=Ikev2EncryptionAlgorithm.ENCR_AES_CBC,
            key_length=128
        )
        transform_encr.next_payload = TransformNextPayload.LAST

        proposal = Ikev2Proposal(
            protocol_id=Ikev2ProtocolId.IKE,
            transforms=[transform_prf, transform_dh, transform_encr],
            spi=bytes()
        )

        sa_payload = Ikev2PayloadSecurityAssociation(
            flags=set(),
            proposals=[proposal]
        )

        found_prf = sa_payload.get_transform_by_type(Ikev2TransformType.PRF)
        self.assertEqual(found_prf, transform_prf)
        self.assertEqual(found_prf.get_transform_type(), Ikev2TransformType.PRF)

        found_dh = sa_payload.get_transform_by_type(Ikev2TransformType.DH)
        self.assertEqual(found_dh, transform_dh)
        self.assertEqual(found_dh.get_transform_type(), Ikev2TransformType.DH)

        found_encr = sa_payload.get_transform_by_type(Ikev2TransformType.ENCR)
        self.assertEqual(found_encr, transform_encr)
        self.assertEqual(found_encr.get_transform_type(), Ikev2TransformType.ENCR)

        transform_integ = Ikev2TransformIntegrity(
            transform_id=Ikev2IntegrityAlgorithm.AUTH_HMAC_SHA1_96
        )
        transform_integ.next_payload = TransformNextPayload.LAST

        proposal_with_integ = Ikev2Proposal(
            protocol_id=Ikev2ProtocolId.IKE,
            transforms=[transform_integ],
            spi=bytes()
        )

        sa_payload_with_integ = Ikev2PayloadSecurityAssociation(
            flags=set(),
            proposals=[proposal_with_integ]
        )

        found_integ = sa_payload_with_integ.get_transform_by_type(Ikev2TransformType.INTEG)
        self.assertEqual(found_integ, transform_integ)
        self.assertEqual(found_integ.get_transform_type(), Ikev2TransformType.INTEG)

    def test_get_transform_by_type_multiple_proposals(self):
        transform_prf1 = Ikev2TransformPrf(
            transform_id=Ikev2PseudorandomFunction.PRF_HMAC_SHA1
        )
        transform_prf1.next_payload = TransformNextPayload.LAST

        transform_prf2 = Ikev2TransformPrf(
            transform_id=Ikev2PseudorandomFunction.PRF_HMAC_SHA2_256
        )
        transform_prf2.next_payload = TransformNextPayload.LAST

        proposal1 = Ikev2Proposal(
            protocol_id=Ikev2ProtocolId.IKE,
            transforms=[transform_prf1],
            spi=bytes()
        )

        proposal2 = Ikev2Proposal(
            protocol_id=Ikev2ProtocolId.IKE,
            transforms=[transform_prf2],
            spi=bytes()
        )

        sa_payload = Ikev2PayloadSecurityAssociation(
            flags=set(),
            proposals=[proposal1, proposal2]
        )

        found_prf = sa_payload.get_transform_by_type(Ikev2TransformType.PRF)
        self.assertEqual(found_prf.get_transform_type(), Ikev2TransformType.PRF)
        self.assertIn(found_prf, [transform_prf1, transform_prf2])

    def test_get_transform_by_type_not_found(self):
        sa_payload = Ikev2PayloadSecurityAssociation(
            flags=set(),
            proposals=[self.proposal_single]
        )

        with self.assertRaises(KeyError) as context_manager:
            sa_payload.get_transform_by_type(Ikev2TransformType.DH)
        self.assertEqual(context_manager.exception.args[0], Ikev2TransformType.DH)

    def test_get_transform_by_type_empty_proposals(self):
        sa_payload = Ikev2PayloadSecurityAssociation(
            flags=set(),
            proposals=[]
        )

        with self.assertRaises(KeyError) as context_manager:
            sa_payload.get_transform_by_type(Ikev2TransformType.PRF)
        self.assertEqual(context_manager.exception.args[0], Ikev2TransformType.PRF)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import collections
import enum

import attr

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.base import VariantParsable
from cryptoparser.common.exception import InvalidType, NotEnoughData
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary


class OpenVpnOpCode(enum.IntEnum):
    CONTROL_V1 = 0x04
    ACK_V1 = 0x05
    HARD_RESET_CLIENT_V2 = 0x07
    HARD_RESET_SERVER_V2 = 0x08


class OpenVpnPacketWrapperTcp(ParsableBase):
    def __init__(self, payload):
        self.payload = payload

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)
        parser.parse_bytes('payload', 2)

        return OpenVpnPacketWrapperTcp(parser['payload']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_bytes(self.payload, 2)

        return composer.composed_bytes


@attr.s
class OpenVpnPacketBase(ParsableBase):
    HEADER_SIZE = 10

    session_id = attr.ib(validator=attr.validators.instance_of(int))
    packet_id_array = attr.ib(
        validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(int))
    )
    remote_session_id = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))

    @classmethod
    @abc.abstractmethod
    def get_op_code(cls):
        return NotImplementedError()  # pragma: no cover

    @classmethod
    def parse_header(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('packet_type', 1)
        if parser['packet_type'] >> 3 != cls.get_op_code():
            raise InvalidType()

        parser.parse_numeric('session_id', 8)
        parser.parse_numeric('packet_id_array_length', 1)
        if parser['packet_id_array_length']:
            parser.parse_numeric_array('packet_id_array', parser['packet_id_array_length'], 4)
            parser.parse_numeric('remote_session_id', 8)

            packet_id_array = parser['packet_id_array']
            remote_session_id = parser['remote_session_id']
        else:
            packet_id_array = []
            remote_session_id = None

        return parser['session_id'], packet_id_array, remote_session_id, parser.parsed_length

    def _compose_header(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.get_op_code() << 3, 1)
        composer.compose_numeric(self.session_id, 8)
        composer.compose_numeric(len(self.packet_id_array), 1)
        if self.packet_id_array:
            composer.compose_numeric_array(self.packet_id_array, 4)
            composer.compose_numeric(self.remote_session_id, 8)

        return composer.composed_bytes


@attr.s
class OpenVpnPacketControlV1(OpenVpnPacketBase):
    packet_id = attr.ib(validator=attr.validators.instance_of(int))
    payload = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def get_op_code(cls):
        return OpenVpnOpCode.CONTROL_V1

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.packet_id, 4)
        composer.compose_raw(self.payload)

        return self._compose_header() + composer.composed_bytes

    @classmethod
    def _parse(cls, parsable):
        session_id, packet_id_array, remote_session_id, header_length = cls.parse_header(parsable)

        body_parser = ParserBinary(parsable[header_length:])
        body_parser.parse_numeric('packet_id', 4)
        body_parser.parse_raw('payload', body_parser.unparsed_length)

        return OpenVpnPacketControlV1(
            session_id,
            packet_id_array,
            remote_session_id,
            body_parser['packet_id'],
            body_parser['payload']
        ), header_length + body_parser.parsed_length


@attr.s(init=False)
class OpenVpnPacketAckV1(OpenVpnPacketBase):
    def __init__(self, session_id, remote_session_id, packet_id_array):
        super().__init__(session_id, packet_id_array, remote_session_id)

    @classmethod
    def get_op_code(cls):
        return OpenVpnOpCode.ACK_V1

    def compose(self):
        return self._compose_header()

    @classmethod
    def _parse(cls, parsable):
        session_id, packet_id_array, remote_session_id, header_length = cls.parse_header(parsable)

        return OpenVpnPacketAckV1(session_id, remote_session_id, packet_id_array), header_length


@attr.s(init=False)
class OpenVpnPacketHardResetClientV2(OpenVpnPacketBase):
    def __init__(self, session_id, packet_id):
        super().__init__(session_id, packet_id_array=[], remote_session_id=None)

        self.packet_id = packet_id

    @classmethod
    def get_op_code(cls):
        return OpenVpnOpCode.HARD_RESET_CLIENT_V2

    @classmethod
    def _parse(cls, parsable):
        session_id, packet_id_array, _, header_length = cls.parse_header(parsable)
        if packet_id_array:
            raise InvalidValue(packet_id_array, cls, 'packet_id_array')

        body_parser = ParserBinary(parsable[header_length:])
        body_parser.parse_numeric('packet_id', 4)

        return OpenVpnPacketHardResetClientV2(
            session_id, body_parser['packet_id']
        ), header_length + body_parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.packet_id, 4)

        return self._compose_header() + composer.composed_bytes


@attr.s(init=False)
class OpenVpnPacketHardResetServerV2(OpenVpnPacketBase):
    def __init__(self, session_id, remote_session_id, packet_id_array, packet_id):
        super().__init__(session_id, packet_id_array, remote_session_id)

        self.packet_id = packet_id

    @classmethod
    def get_op_code(cls):
        return OpenVpnOpCode.HARD_RESET_SERVER_V2

    @classmethod
    def _parse(cls, parsable):
        session_id, packet_id_array, remote_session_id, header_length = cls.parse_header(parsable)

        body_parser = ParserBinary(parsable[header_length:])
        body_parser.parse_numeric('packet_id', 4)

        return OpenVpnPacketHardResetServerV2(
            session_id,
            remote_session_id,
            packet_id_array,
            body_parser['packet_id']
        ), header_length + body_parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.packet_id, 4)

        return self._compose_header() + composer.composed_bytes


class OpenVpnPacketVariant(VariantParsable):
    _VARIANTS = collections.OrderedDict([
        (OpenVpnOpCode.ACK_V1, [OpenVpnPacketAckV1, ]),
        (OpenVpnOpCode.CONTROL_V1, [OpenVpnPacketControlV1, ]),
        (OpenVpnOpCode.HARD_RESET_CLIENT_V2, [OpenVpnPacketHardResetClientV2, ]),
        (OpenVpnOpCode.HARD_RESET_SERVER_V2, [OpenVpnPacketHardResetServerV2, ]),
    ])

    @classmethod
    def _get_variants(cls):
        return cls._VARIANTS
